from typing import Union
from numbers import Number
import random
import math
import numpy as np

from .._utils_1d import inverse_covariance
from .._state import State, DataNoiseState


class Target:
    """Observed data with noise that can be treated as an unknown

    Parameters
    ----------
    name : str
        name of the data, for display purposes
    dobs : np.ndarray
        numerical data
    covariance_mat_inv : Union[Number, np.ndarray], optional
        the inverse of the data covariance matrix, either a number or a full matrix, by
        default None
    noise_is_correlated : bool, optional
        whether the noise between data points is correlated or not, by default False
    std_min : Number, optional
        the minimum value of the standard deviation of data noise, by default 0.01
    std_max : Number, optional
        the maximum value of the standard deviation of data noise, by default 1
    std_perturb_std : Number, optional
        the perturbation standard deviation of the standard deviation of data noise, by
        default 0.1
    correlation_min : Number, optional
        the miminum value of the correlation of data noise, by default 0.01
    correlation_max : Number, optional
        the maximum value of the correlation of data noise, by default 1
    correlation_perturb_std : Number, optional
        the perturbation standard deviation of the standard deviation of data noise, by
        default 0.1
    """

    def __init__(
        self,
        name: str,
        dobs: np.ndarray,
        covariance_mat_inv: Union[Number, np.ndarray] = None,
        noise_is_correlated: bool = False,
        std_min: Number = 0.01,
        std_max: Number = 1,
        std_perturb_std: Number = 0.1,
        correlation_min: Number = 0.01,
        correlation_max: Number = 1,
        correlation_perturb_std: Number = 0.1,
    ):
        self._name = name
        self.dobs = np.array(dobs)
        self.noise_is_correlated = noise_is_correlated
        self.std_min = std_min
        self.std_max = std_max
        self.std_perturb_std = std_perturb_std
        self.correlation_min = correlation_min
        self.correlation_max = correlation_max
        self.correlation_perturb_std = correlation_perturb_std
        assert (
            std_min <= std_max
        ), "standard deviation minimum should be less than maximum"
        assert std_min >= 0, "standard deviation should always be positive"
        assert (
            correlation_min <= correlation_max
        ), "correlation minimum should be less than maximum"
        assert correlation_min >= 0, "correlation should always be positive"
        assert correlation_max <= 1, "correlation should always be less than 1"
        if covariance_mat_inv is not None:
            self._perturbation_func = None
            if np.isscalar(covariance_mat_inv):
                self.covariance_mat_inv = covariance_mat_inv
            else:
                self.covariance_mat_inv = np.array(covariance_mat_inv)
        else:
            self.covariance_mat_inv = None

    def __repr__(self):
        _repr_args = [
            f"name='{self.name}",
            f"dobs=ndarray of shape {self.dobs.shape}",
            f"is_hierarchical={self.is_hierarchical}",
        ]
        if not self.is_hierarchical:
            _str = (
                f"ndarray of shape {self.covariance_mat_inv.shape}"
                if isinstance(self.covariance_mat_inv, np.ndarray)
                else self.covariance_mat_inv
            )
            _repr_args.append(f"covariance_mat_inv={_str}")
        else:
            for k, v in vars(self).items():
                if (
                    not k.startswith("_")
                    and k not in ["dobs", "covariance_mat_inv"]
                    and v is not None
                ):
                    if not self.noise_is_correlated and k.startswith("correlation"):
                        continue
                    _repr_args.append(f"{k}={v}")
        return f"{self.__class__.__name__}({', '.join(_repr_args)})"

    @property
    def name(self) -> str:
        return self._name

    @property
    def is_hierarchical(self):
        """whether the data noise is unknown (i.e. to be inverted for)"""
        return self.covariance_mat_inv is None

    def initialize(self, state: State):
        """initializes the data noise parameters

        Parameters
        ----------
        state : State
            the current state in the Bayesian inference, in which DataNoiseState
            is to be set
        """
        if self.is_hierarchical:
            noise_std = random.uniform(self.std_min, self.std_max)
            # state.set_param_values((self.name, "noise_std"), noise_std)
            noise_corr = (
                random.uniform(self.correlation_min, self.correlation_max)
                if self.noise_is_correlated
                else None
            )
            state.set_param_values(
                self.name, DataNoiseState(std=noise_std, correlation=noise_corr)
            )

    def inverse_covariance_times_vector(
        self, state: State, vector: np.ndarray
    ) -> np.ndarray:
        """calculates the dot product of the covariance inverse matrix with a given
        vector

        Parameters
        ----------
        state : State
            the current state state
        vector : np.ndarray
            the vector to apply the dot product on

        Returns
        -------
        np.ndarray
            the result from the dot product operation
        """
        if not self.is_hierarchical:
            if np.isscalar(self.covariance_mat_inv):
                return self.covariance_mat_inv * vector
            else:
                if self.covariance_mat_inv.ndim == 2:
                    return self.covariance_mat_inv @ vector
                return self.covariance_mat_inv * vector
        else:
            noise = state[self.name]
            std = noise.std
            correlation = noise.correlation
            if correlation is None:
                return 1 / std**2 * vector
            else:
                n = self.dobs.size
                mat = inverse_covariance(std, correlation, n)
                return mat @ vector

    def log_determinant_covariance(self, state: State) -> float:
        r"""the log of the determinant of the covariance matrix

        The determinant of the data covariance matrix is calculated assuming
        an exponential decay in the noise correlation between adjacent data
        points [1]_, i.e.,

        .. math::
            \lvert \mathbf{C}_e \rvert = \sigma^{2n} (1 - r^2)^{n-1},

        where :math:`\sigma` denotes the standard deviation, :math:`r` the
        correlation, and :math:`n` the size of the data vector.


        Parameters
        ----------
        state : State
            the current Bayesian inferernce state

        Returns
        -------
        float
            the log of the determinant

        References
        ----------
        .. [1] Bodin et al. 2012, Transdimensional inversion of receiver functions
            and surface wave dispersion.
        """
        noise = state[self.name]
        std = noise.std
        r = noise.correlation
        if r is None:
            r = 0
        n = self.dobs.size
        log_det = (2 * n) * math.log(std) + (n - 1) * math.log(1 - r**2)
        return log_det



from bisect import bisect_left
import math
from typing import Tuple, Union, List, Dict, Callable
from numbers import Number
import random
import numpy as np
import scipy.spatial
import shapely.geometry
import matplotlib as mpl
import matplotlib.pyplot as plt

from ..parameterization._parameter_space import ParameterSpace
from ..perturbations._param_values import ParamPerturbation
from ..perturbations._param_space import ParamSpacePerturbation
from ..perturbations._birth_death import BirthPerturbation, DeathPerturbation
from ._discretization import Discretization
from ..prior import Prior
from .._state import State, ParameterSpaceState
from .._utils_1d import (
    interpolate_depth_profile,
    interpolate_nearest_1d,
    compute_voronoi1d_cell_extents,
    insert_1d,
    delete_1d,
    nearest_neighbour_1d,
)


SQRT_TWO_PI = math.sqrt(2 * math.pi)


def _plot(x, y, ax, swap_xy_axes=False, **kwargs):
    if swap_xy_axes:
        ax.plot(y, x, **kwargs)
    else:
        ax.plot(x, y, **kwargs)


class Voronoi(Discretization):
    r"""Utility class for Voronoi tessellation

    Parameters
    ----------
    name : str
        name attributed to the Voronoi tessellation, for display and storing
        purposes
    spatial_dimensions : int
        number of dimensions of the desired Voronoi tessellation, e.g. 1D,
        2D, or 3D.
    vmin, vmax : Union[Number, np.ndarray]
        minimum/maximum value bounding each dimension
    perturb_std : Union[Number, np.ndarray]
        standard deviation of the Gaussians used to randomly perturb the Voronoi
        sites in each dimension.
    n_dimensions : Number, optional
        number of dimensions. None (default) results in a trans-dimensional
        discretization, with the dimensionality of the parameter space allowed
        to vary in the range ``n_dimensions_min``-``n_dimensions_max``
    n_dimensions_min, n_dimensions_max : Number, optional
        minimum and maximum number of dimensions, by default 1 and 10. These
        parameters are ignored if ``n_dimensions`` is not None, i.e. if the
        discretization is not trans-dimensional
    n_dimensions_init_range : Number, optional
        percentage of the range `n_dimensions_min`` - ``n_dimensions_max`` used to
        initialize the number of dimensions (0.3. by default). For example, if
        ``n_dimensions_min`` = 1, ``n_dimensions_max`` = 10, and
        ``n_dimensions_init_range`` = 0.5,
        the maximum number of dimensions at the initialization is

            int((n_dimensions_max - n_dimensions_min) * n_dimensions_init_range + n_dimensions_max)

    parameters : List[Prior], optional
        a list of free parameters, by default None
    birth_from : {"prior", "neighbour"}, optional
        whether to initialize the free parameters associated with the newborn
        Voronoi cell by randomly drawing from their prior or by perturbing the
        value found in the nearest Voronoi cell (default).
    """

    def __init__(
        self,
        name: str,
        spatial_dimensions: Number,
        vmin: Union[Number, np.ndarray],
        vmax: Union[Number, np.ndarray],
        perturb_std: Union[Number, np.ndarray],
        n_dimensions: int = None,
        n_dimensions_min: int = 2,
        n_dimensions_max: int = 10,
        n_dimensions_init_range: Number = 0.3,
        parameters: List[Prior] = None,
        birth_from: str = "neighbour",  # either "neighbour" or "prior"
    ):
        super().__init__(
            name=name,
            spatial_dimensions=spatial_dimensions,
            perturb_std=perturb_std,
            n_dimensions=n_dimensions,
            n_dimensions_min=n_dimensions_min,
            n_dimensions_max=n_dimensions_max,
            n_dimensions_init_range=n_dimensions_init_range,
            parameters=parameters,
            birth_from=birth_from,
            vmin=vmin,
            vmax=vmax,
        )
        self.vmin = vmin
        self.vmax = vmax
        msg = "The %s number of Voronoi cells, "
        if n_dimensions is not None:
            assert n_dimensions > 0, (
                msg % "minimum" + "`n_dimensions`, should be greater than zero"
            )
            assert isinstance(n_dimensions, int), (
                msg % "minimum" + "`n_dimensions`, should be an integer"
            )
            assert isinstance(n_dimensions, int), (
                msg % "maximum" + "`n_dimensions`, should be an integer"
            )

    def sample_site(self) -> np.ndarray:
        """draws a Voronoi-site position at random within the discretization domain"""
        return np.random.uniform(self.vmin, self.vmax, self.spatial_dimensions)

    def sample_discretization(self) -> ParameterSpaceState:
        # initialize number of dimensions
        if not self.trans_d:
            n_voronoi_cells = self._n_dimensions
        else:
            n_dims_min = self._n_dimensions_min
            n_dims_max = self._n_dimensions_max
            n_voronoi_cells = random.randint(n_dims_min, n_dims_max)

        # initialize Voronoi sites
        voronoi_sites = np.array([self.sample_site() for _ in range(n_voronoi_cells)])
        if self.spatial_dimensions == 1:
            voronoi_sites = np.sort(np.ravel(voronoi_sites))

        # initialize parameter values
        parameter_vals = {"discretization": voronoi_sites}
        return ParameterSpaceState(n_voronoi_cells, parameter_vals)

    def initialize(
        self, position: np.ndarray = None
    ) -> Union[ParameterSpaceState, List[ParameterSpaceState]]:
        """initializes the parameter space linked to the Voronoi tessellation

        Returns
        -------
        Union[ParameterSpaceState, List[ParameterSpaceState]
            an initial parameter space state, or a list of parameter space states
        """
        if position is None:
            return self._initialize()
        else:
            return [self._initialize() for _ in position]

    def _initialize(self) -> ParameterSpaceState:
        # initialize number of dimensions
        if not self.trans_d:
            n_voronoi_cells = self._n_dimensions
        else:
            init_range = self._n_dimensions_init_range
            n_dims_min = self._n_dimensions_min
            n_dims_max = self._n_dimensions_max
            init_max = int((n_dims_max - n_dims_min) * init_range + n_dims_min)
            n_voronoi_cells = random.randint(n_dims_min, init_max)

        # initialize Voronoi sites
        voronoi_sites = np.array([self.sample_site() for _ in range(n_voronoi_cells)])
        if self.spatial_dimensions == 1:
            voronoi_sites = np.sort(np.ravel(voronoi_sites))

        # initialize parameter values
        parameter_vals = {"discretization": voronoi_sites}
        for name, param in self.parameters.items():
            parameter_vals[name] = param.initialize(voronoi_sites)
        return ParameterSpaceState(n_voronoi_cells, parameter_vals)

    def _perturb_site(
        self, site: Union[Number, np.ndarray]
    ) -> Union[Number, np.ndarray]:
        """perturbes a Voronoi  site

        Parameters
        ----------
        site : Union[Number, np.ndarray]
            Voronoi site position.

        Returns
        -------
        Union[Number, np.ndarray]
            perturbed Voronoi site position
        """
        while True:
            if self.spatial_dimensions == 1:
                random_deviate = random.normalvariate(0, self.perturb_std)
                new_site = site + random_deviate
                if new_site >= self.vmin and new_site <= self.vmax:
                    return new_site
            else:
                random_deviate = np.random.normal(
                    0, self.perturb_std, self.spatial_dimensions
                )
                new_site = site + random_deviate
                if all((new_site >= self.vmin) & (new_site <= self.vmax)):
                    return new_site

    def perturb_value(
        self, old_ps_state: ParameterSpaceState, isite: int
    ) -> Tuple[ParameterSpaceState, Number]:
        r"""perturbs the value of one Voronoi site and calculates the log of the
        partial acceptance probability

        .. math::
            \underbrace{\alpha_{p}}_{\begin{array}{c} \text{Partial} \\ \text{acceptance} \\ \text{probability} \end{array}} =
            \underbrace{\frac{p\left({\bf m'}\right)}{p\left({\bf m}\right)}}_{\text{Prior ratio}}
            \underbrace{\frac{q\left({\bf m} \mid {\bf m'}\right)}{q\left({\bf m'} \mid {\bf m}\right)}}_{\text{Proposal ratio}}
            \underbrace{\lvert \mathbf{J} \rvert}_{\begin{array}{c} \text{Jacobian} \\ \text{determinant} \end{array}}.

        Parameters
        ----------
        old_ps_state : ParameterSpaceState
            the current parameter space state
        isite : Number
            the index of the Voronoi site to be perturbed

        Returns
        -------
        Tuple[ParameterSpaceState, Number]
            the new parameter space state and its associated partial acceptance
            probability excluding log likelihood ratio
        """
        old_sites = old_ps_state["discretization"]
        old_site = old_sites[isite]
        new_site = self._perturb_site(old_sites[isite])
        new_sites = old_sites.copy()
        new_sites[isite] = new_site
        if self.spatial_dimensions == 1:
            isort = np.argsort(new_sites)
            new_sites = new_sites[isort]
            new_values = {"discretization": new_sites}
            log_prior_ratio = 0
            for param_name, param in self.parameters.items():
                values = old_ps_state[param_name]
                if not isinstance(param, ParameterSpace) and param.position is not None:
                    log_prior_old = param.log_prior(values[isite], old_site)
                    log_prior_new = param.log_prior(values[isite], new_site)
                    log_prior_ratio += log_prior_new - log_prior_old
                new_values[param_name] = (
                    values[isort]
                    if isinstance(values, np.ndarray)
                    else [values[i] for i in isort]
                )
        else:
            new_values = {"discretization": new_sites}
            log_prior_ratio = 0
            for param_name, param in self.parameters.items():
                values = old_ps_state[param_name]
                if not isinstance(param, ParameterSpace) and param.position is not None:
                    log_prior_old = param.log_prior(values[isite], old_site)
                    log_prior_new = param.log_prior(values[isite], new_site)
                    log_prior_ratio += log_prior_new - log_prior_old
                new_values[param_name] = values

        new_ps_state = ParameterSpaceState(old_ps_state.n_dimensions, new_values)
        return (
            new_ps_state,
            log_prior_ratio,
        )  # log_proposal_ratio=0 and log_det_jacobian=0

    def birth(
        self, old_ps_state: ParameterSpaceState
    ) -> Tuple[ParameterSpaceState, float]:
        r"""creates a new Voronoi cell, initializes all free parameters
        associated with it, and returns the pertubed state along with the
        log of the corresponding partial acceptance probability,

        .. math::
            \underbrace{\alpha_{p}}_{\begin{array}{c} \text{Partial} \\ \text{acceptance} \\ \text{probability} \end{array}} =
            \underbrace{\frac{p\left({\bf m'}\right)}{p\left({\bf m}\right)}}_{\text{Prior ratio}}
            \underbrace{\frac{q\left({\bf m} \mid {\bf m'}\right)}{q\left({\bf m'} \mid {\bf m}\right)}}_{\text{Proposal ratio}}
            \underbrace{\lvert \mathbf{J} \rvert}_{\begin{array}{c} \text{Jacobian} \\ \text{determinant} \end{array}}.

        In this case, the prior probability of the model :math:`{\bf m}` is

        .. math::
            p({\bf m}) = p({\bf c} \mid k) p(k) \prod_i{p({\bf v}_i \mid {\bf c})} ,

        where :math:`k` denotes the number of Voronoi cells, each entry of the
        vector :math:`{\bf c}` corresponds to the position of a Voronoi site,
        and each :math:`i`\ th free parameter :math:`{\bf v}` has the same
        dimensionality as :math:`{\bf c}`.

        Following [1]_, :math:`p({\bf c} \mid k) = \frac{k! \left(N - k \right)!}{N!}`. If we then
        assume that :math:`p(k) = \frac{1}{\Delta k}`, where :math:`\Delta k = k_{max} - k_{min}`,
        the prior ratio reads

        .. math::
            \frac{p({\bf m'})}{p({\bf m})} =
            \frac{(k+1) \prod_i p(v_i^{k+1})}{(N-k)},

        where :math:`p(v_i^{k+1})` denotes the prior probability of the newly
        born :math:`i`\ th parameter, which may be dependent on :math:`{\bf c}`.
        The proposal ratio reads

        .. math::
            \frac{q({\bf m} \mid {\bf m'})}{q({\bf m'} \mid {\bf m})} =
            \frac{(N-k)}{(k+1) \prod_i q_{v_i}^{k+1}},

        where :math:`q_{v_i}^{k+1}` denotes the proposal probability for the
        newly born :math:`i`\ th parameter in the new dimension. It is easy to
        show that, in the case of a birth from neighbor [1]_ or a birth from
        prior [2]_ (see :attr:`birth_from`), :math:`\lvert \mathbf{J} \rvert = 1`
        and :math:`\alpha_{p} = \frac{p({\bf m'})}{p({\bf m})} \frac{q({\bf m} \mid {\bf m'})}{q({\bf m'} \mid {\bf m})}`.
        It follows that

        .. math::
            \alpha_{p} =
            \frac{(k+1) \prod_i p(v_i^{k+1})}{(N-k)} \frac{(N-k)}{(k+1) \prod_i q_{v_i}^{k+1}} =
            \frac{\prod_i p(v_i^{k+1})}{\prod_i{q_{v_i}^{k+1}}}.

        In the case of a birth from prior, :math:`q_{v_i}^{k+1} = p(v_i^{k+1})`
        and

        .. math::
            \alpha_{p} =
            \frac{\prod_i p(v_i^{k+1})}{\prod_i{p(v_i^{k+1})}} = 1.

        In the case of a birth from neighbor, :math:`q_{v_i}^{k+1} =
        \frac{1}{\theta \sqrt{2 \pi}} \exp \lbrace -\frac{\left( v_i^{k+1} - v_i \right)^2}{2\theta^2} \rbrace`,
        where the newly born value, :math:`v_i^{k+1}`, is generated by perturbing
        the original value, :math:`v_i`, of the :math:`i`\ th parameter. This is
        achieved through a random deviate from the normal distribution
        :math:`\mathcal{N}(v_i, \theta)`, with :math:`\theta` denoting the
        standard deviation of the Gaussian used to carry out the perturbation
        (see, for example, :attr:`bayesbay.prior.UniformPrior.perturb_std`) .
        The partial acceptance probability is then computed numerically.


        Parameters
        ----------
        old_ps_state : ParameterSpaceState
            current parameter space state

        Returns
        -------
        ParameterSpaceState
            new parameter space state
        Number
            log of the partial acceptance probability,
            :math:`log(\alpha_{p}) = \log(\frac{\prod_i p(v_i^{k+1})}{\prod_i{q_{v_i}^{k+1}}})`

        References
        ----------
        .. [1] Bodin et al. 2012, Transdimensional inversion of receiver functions
            and surface wave dispersion
        .. [2] Hawkins and Sambridge 2015, Geophysical imaging using trans-dimensional
            trees
        """
        # prepare for birth perturbation
        n_cells = old_ps_state.n_dimensions
        if n_cells == self._n_dimensions_max:
            return old_ps_state, -math.inf
        # randomly choose a new Voronoi site position
        new_site = self.sample_site()
        old_sites = old_ps_state["discretization"]
        initialized_values, log_prob_ratio = self._initialize_newborn_params(
            new_site, old_sites, old_ps_state
        )
        new_values = dict()
        if self.spatial_dimensions == 1:
            idx_insert = bisect_left(old_sites, new_site)
            new_sites = insert_1d(old_sites, idx_insert, new_site)
            new_values["discretization"] = new_sites
            for name, value in initialized_values.items():
                old_values = old_ps_state[name]
                if isinstance(old_values, np.ndarray):
                    new_values[name] = insert_1d(old_values, idx_insert, value)
                else:
                    new_values[name] = (
                        old_values[:idx_insert] + [value] + old_values[idx_insert:]
                    )
        else:
            idx_insert = n_cells
            new_sites = np.row_stack((old_sites, new_site))
            new_values["discretization"] = new_sites
            for name, value in initialized_values.items():
                old_values = old_ps_state[name]
                if isinstance(old_values, np.ndarray):
                    new_values[name] = np.append(old_values, value)
                else:
                    new_values[name] = old_values + [value]
        new_ps_state = ParameterSpaceState(n_cells + 1, new_values)
        return new_ps_state, log_prob_ratio

    def death(self, old_ps_state: ParameterSpaceState):
        r"""removes a new Voronoi cell and returns the pertubed state along with
        the log of the corresponding partial acceptance probability,

        .. math::
            \underbrace{\alpha_{p}}_{\begin{array}{c} \text{Partial} \\ \text{acceptance} \\ \text{probability} \end{array}} =
            \underbrace{\frac{p\left({\bf m'}\right)}{p\left({\bf m}\right)}}_{\text{Prior ratio}}
            \underbrace{\frac{q\left({\bf m} \mid {\bf m'}\right)}{q\left({\bf m'} \mid {\bf m}\right)}}_{\text{Proposal ratio}}
            \underbrace{\lvert \mathbf{J} \rvert}_{\begin{array}{c} \text{Jacobian} \\ \text{determinant} \end{array}}.

        It is straightforward to show that this equals the reciprocal of
        the partial acceptance probability obtained in the case of a birth
        perturbation (see :meth:`birth`), i.e.,

        .. math::
            \alpha_{p} = \frac{\prod_i{q_{v_i}^{k+1}}}{\prod_i p(v_i^{k+1})}.

        Parameters
        ----------
        old_ps_state : ParameterSpaceState
            current parameter space state

        Returns
        -------
        ParameterSpaceState
            new parameter space state
        Number
            log of the partial acceptance probability,
            :math:`log(\alpha_{p}) = -\log(\frac{\prod_i p(v_i^{k+1})}{\prod_i{q_{v_i}^{k+1}}})`
        """
        # prepare for death perturbation
        n_cells = old_ps_state.n_dimensions
        if n_cells == self._n_dimensions_min:
            return old_ps_state, -math.inf
        # randomly choose an existing Voronoi site to kill
        iremove = random.randint(0, n_cells - 1)
        # remove parameter values for the removed site
        new_values = dict()
        for name, old_values in old_ps_state.param_values.items():
            if isinstance(old_values, np.ndarray):  # pure Prior
                if self.spatial_dimensions == 1:
                    new_values[name] = delete_1d(old_values, iremove)
                else:
                    new_values[name] = np.delete(old_values, iremove, axis=0)
            else:  # ParameterSpace
                new_values[name] = old_values[:iremove] + old_values[iremove + 1 :]
        new_ps_state = ParameterSpaceState(n_cells - 1, new_values)
        return new_ps_state, self._log_prob_death_parameters(
            old_ps_state, new_ps_state, iremove
        )

    def log_prior(self, *args):
        r"""
        BayesBay implements the grid trick, which calculates the prior
        probability of a Voronoi discretization through the combinatorial
        formula :math:`{N \choose k}^{-1}`, with `k` denoting the number of
        Voronoi sites and `N` the number of possible positions allowed for the
        sites [3]_.

        References
        ----------
        .. [3] Bodin and Sambridge (2009), Seismic tomography with the reversible
            jump algorithm
        """
        raise NotImplementedError

    def _init_perturbation_funcs(self):
        self._perturbation_funcs = []
        self._perturbation_weights = []
        _ps_perturbation_funcs = []
        _ps_perturbation_weights = []
        if self.trans_d:
            _ps_perturbation_funcs.append(BirthPerturbation(self))
            _ps_perturbation_funcs.append(DeathPerturbation(self))
            _ps_perturbation_weights.append(1)
            _ps_perturbation_weights.append(1)
        if self.parameters:
            # initialize parameter values perturbation
            _params = self.parameters.values()
            _prior_pars = [p for p in _params if not isinstance(p, ParameterSpace)]
            if _prior_pars:
                _ps_perturbation_funcs.append(ParamPerturbation(self.name, _prior_pars))
                _ps_perturbation_weights.append(3)
            # initialize nested parameter space perturbations
            _ps_pars = [p for p in _params if isinstance(p, ParameterSpace)]
            for ps in _ps_pars:
                _funcs = ps.perturbation_funcs
                self._perturbation_funcs.extend(_funcs)
                self._perturbation_weights.extend(ps.perturbation_weights)
        _ps_perturbation_funcs.append(ParamPerturbation(self.name, [self]))
        _ps_perturbation_weights.append(1)
        self._perturbation_funcs.append(
            ParamSpacePerturbation(
                self.name, _ps_perturbation_funcs, _ps_perturbation_weights
            )
        )
        self._perturbation_weights.append(sum(_ps_perturbation_weights))

    @property
    def perturbation_funcs(self) -> List[Callable[[State], Tuple[State, Number]]]:
        r"""the list of perturbation functions allowed in the parameter space linked to
        the Voronoi discretization. Each function takes in a state (see :class:`State`)
        and returns a new state along with the corresponding partial acceptance
        probability,

        .. math::
            \underbrace{\alpha_{p}}_{\begin{array}{c} \text{Partial} \\ \text{acceptance} \\ \text{probability} \end{array}} =
            \underbrace{\frac{p\left({\bf m'}\right)}{p\left({\bf m}\right)}}_{\text{Prior ratio}}
            \underbrace{\frac{q\left({\bf m} \mid {\bf m'}\right)}{q\left({\bf m'} \mid {\bf m}\right)}}_{\text{Proposal ratio}}
            \underbrace{\lvert \mathbf{J} \rvert}_{\begin{array}{c} \text{Jacobian} \\ \text{determinant} \end{array}},

        """
        return self._perturbation_funcs

    @property
    def perturbation_weights(self) -> List[Number]:
        """a list of perturbation weights, corresponding to each of the
        :meth:`perturbation_funcs` that determines the probability of each of them
        to be chosen during each step

        The weights are not normalized and have the following default values:

        - Birth/Death perturbations: 1
        - Parameter values perturbation: 3
        - Voronoi site perturbation: 1
        """
        return self._perturbation_weights

    def nearest_neighbour(
        self, discretization: np.ndarray, query_point: Union[Number, np.ndarray]
    ) -> Union[Number, np.ndarray]:
        if self.spatial_dimensions == 1:
            return nearest_neighbour_1d(
                xp=query_point, x=discretization, xlen=discretization.size
            )
        return np.argmin(np.linalg.norm(discretization - query_point, axis=1))

    def log_prob_initialize_discretization(
        self, ps_state: ParameterSpaceState
    ) -> Number:
        return 0


class Voronoi1D(Voronoi):
    r"""Utility class for Voronoi tessellation in 1D

    Parameters
    ----------
    name : str
        name attributed to the Voronoi tessellation, for display and storing
        purposes
    vmin, vmax : Union[Number, np.ndarray]
        minimum/maximum value bounding each dimension
    perturb_std : Union[Number, np.ndarray]
        standard deviation of the Gaussians used to randomly perturb the Voronoi
        sites in each dimension.
    n_dimensions : Number, optional
        number of dimensions. None (default) results in a trans-dimensional
        discretization, with the dimensionality of the parameter space allowed
        to vary in the range ``n_dimensions_min``-``n_dimensions_max``
    n_dimensions_min, n_dimensions_max : Number, optional
        minimum and maximum number of dimensions, by default 1 and 10. These
        parameters are ignored if ``n_dimensions`` is not None, i.e. if the
        discretization is not trans-dimensional
    n_dimensions_init_range : Number, optional
        percentage of the range ``n_dimensions_min`` - ``n_dimensions_max`` used to
        initialize the number of dimensions (0.3. by default). For example, if
        ``n_dimensions_min`` = 1, ``n_dimensions_max`` = 10, and
        ``n_dimensions_init_range`` = 0.5,
        the maximum number of dimensions at the initialization is::

            int((n_dimensions_max - n_dimensions_min) * n_dimensions_init_range + n_dimensions_max)

    parameters : List[Prior], optional
        a list of free parameters, by default None
    birth_from : {"prior", "neighbour"}, optional
        whether to initialize the free parameters associated with the newborn
        Voronoi cell by randomly drawing from their prior or by perturbing the
        value found in the nearest Voronoi cell (default).
    """

    def __init__(
        self,
        name: str,
        vmin: Number,
        vmax: Number,
        perturb_std: Union[Number, np.ndarray],
        n_dimensions: int = None,
        n_dimensions_min: int = 1,
        n_dimensions_max: int = 10,
        n_dimensions_init_range: Number = 0.3,
        parameters: List[Prior] = None,
        birth_from: str = "neighbour",  # either "neighbour" or "prior"
    ):
        super().__init__(
            name=name,
            spatial_dimensions=1,
            vmin=vmin,
            vmax=vmax,
            perturb_std=perturb_std,
            n_dimensions=n_dimensions,
            n_dimensions_min=n_dimensions_min,
            n_dimensions_max=n_dimensions_max,
            n_dimensions_init_range=n_dimensions_init_range,
            parameters=parameters,
            birth_from=birth_from,
        )

    @staticmethod
    def compute_cell_extents(voronoi_sites: np.ndarray, lb=0, ub=None, fill_value=0):
        r"""compute Voronoi cell extents from the Voronoi sites. Voronoi-cell
        boundaries are first drawn at the midpoint between consecutive Voronoi
        nuclei. The extent is then derived from the distance between consecutive
        boundaries.

        Parameters
        ----------
        voronoi_sites : np.ndarray of shape (n,)
            Voronoi-site positions. These should be greater or equal to zero

        lb, ub : float
            Lower and upper bounds used in the calculation of Voronoi-cell
            extents. `None` values for `lb` or `ub` denote an unbounded cell.
            The extent of an unbounded cell is set to `fill_value`.

        fill_value : float
            Value attributed to unbounded Voronoi cells

        Returns
        -------
        np.ndarray
            Voronoi-cell extents

        Examples
        --------
        >>> voronoi_sites = np.array([2, 5.5, 8, 10])

        >>> Voronoi1D.compute_cell_extents(voronoi_sites, lb=0, ub=None, fill_value=np.nan)
        array([3.75, 3.  , 2.25,  nan])

        >>> Voronoi1D.compute_cell_extents(voronoi_sites, lb=None, ub=None, fill_value=np.nan)
        array([ nan, 3.  , 2.25,  nan])

        >>> Voronoi1D.compute_cell_extents(voronoi_sites, lb=0, ub=15, fill_value=np.nan)
        array([3.75, 3.  , 2.25, 6.  ])
        """
        lb = lb if lb is not None else -np.inf
        ub = ub if ub is not None else np.inf
        return compute_voronoi1d_cell_extents(
            voronoi_sites, lb=lb, ub=ub, fill_value=fill_value
        )

    @staticmethod
    def compute_interface_positions(        
        voronoi_cells: np.ndarray,
        input_type="nuclei",
        lb_tessellation=None,
    ):
        """computes the position of Voronoi-cell interfaces

        Parameters
        ----------
        samples_voronoi_cells : list
            either a list of Voronoi-cell extents or of Voronoi-site positions
            (see ``input_type``)
        input_type : str, {'nuclei', 'extents'}
            argument determining whether each entry of `voronoi_cells` should be
            interpreted as a Voronoi-site position (``'nuclei'``) or as the
            extent of the Voronoi cell (``'extents'``)
        lb_tessellation : Number
            the lower boundary of the 1D tessellation, used to calculate the 
            interface positions when `input_type` is `'extents'`. Ignored otherwise.
            
        Returns
        -------
        np.ndarray
        """
        if input_type == "nuclei":
            return (voronoi_cells[:-1] + voronoi_cells[1:]) / 2
        elif input_type == "extents":
            assert lb_tessellation is not None, (
                "`lb_tessellation` should not be None when `input_type` is"
                "'extents'"
            )
            return np.cumsum(voronoi_cells)[:-1] + lb_tessellation
        else:
            raise ValueError("`input_type` should either be 'nuclei' or 'extents'")
        
    @staticmethod
    def interpolate_tessellation(
        voronoi_cells, param_values, interp_positions, input_type="nuclei"
    ):
        """interpolates the values of a parameter associated with the given
        Voronoi tessellation onto the specified positions

        Parameters
        ----------
        voronoi_cells : np.ndarray
            either Voronoi-cell extents or Voronoi-site positions (see
            ``input_type``)
        param_values : np.ndarray
            the physical parameter value associated with each Voronoi cell
        interp_positions : np.ndarray
            the positions at which the parameter values will be returned
        input_type : str, {'nuclei', 'extents'}
            argument determining whether each entry of `voronoi_cells` should be
            interpreted as a Voronoi-site position (``'nuclei'``) or as the
            extent of the Voronoi cell (``'extents'``)
        Returns
        -------
        np.ndarray
            the physical parameter values associated with ``interp_positions``
        """
        if input_type == "nuclei":
            return interpolate_nearest_1d(interp_positions, voronoi_cells, param_values)
        elif input_type == "extents":
            return interpolate_depth_profile(
                np.array(voronoi_cells), np.array(param_values), interp_positions
            )
        raise ValueError("`input_type` should either be 'nuclei' or 'extents'")

    @staticmethod
    def _interpolate_tessellations(
        samples_voronoi_cells,
        samples_param_values,
        interp_positions,
        input_type="nuclei",
    ):
        interp_params = np.zeros((len(samples_param_values), len(interp_positions)))
        for i, (sample_cells, sample_values) in enumerate(
            zip(samples_voronoi_cells, samples_param_values)
        ):
            interp_params[i, :] = Voronoi1D.interpolate_tessellation(
                np.array(sample_cells),
                np.array(sample_values),
                interp_positions,
                input_type=input_type,
            )
        return interp_params

    @staticmethod
    def get_tessellation_statistics(
        samples_voronoi_cells: list,
        samples_param_values: list,
        interp_positions: np.ndarray,
        percentiles: tuple = (10, 90),
        input_type: str = "nuclei",
    ) -> dict:
        """get the mean, median, std and percentiles of the given ensemble

        Parameters
        ----------
        samples_voronoi_cells : list
            either a list of Voronoi-cell extents or of Voronoi-site positions
            (see ``input_type``)
        samples_param_values : list
            a list of parameter values to draw statistics from
        interp_positions : np.ndarray
            points to interpolate
        percentiles : tuple, optional
            percentiles to calculate, by default (10, 90)
        input_type : str, {'nuclei', 'extents'}
            argument determining whether each entry of `voronoi_cells` should be
            interpreted as a Voronoi-site position (``'nuclei'``) or as the
            extent of the Voronoi cell (``'extents'``)

        Returns
        -------
        dict
            a dictionary with these keys: "mean", "median", "std" and "percentile"
        """
        interp_params = Voronoi1D._interpolate_tessellations(
            samples_voronoi_cells,
            samples_param_values,
            interp_positions,
            input_type=input_type,
        )
        statistics = {
            "mean": np.mean(interp_params, axis=0),
            "median": np.median(interp_params, axis=0),
            "std": np.std(interp_params, axis=0),
            "percentiles": np.percentile(interp_params, percentiles, axis=0),
        }
        return statistics

    @staticmethod
    def get_tessellation_density(
        samples_voronoi_cells: np.ndarray,
        samples_param_values: np.ndarray,
        position_bins: Union[int, np.ndarray] = 100,
        param_value_bins: Union[int, np.ndarray] = 100,
        input_type="nuclei"
    ):
        """plot a 2D density histogram of the Voronoi tessellation

        Parameters
        ----------
        samples_voronoi_cells : list
            either a list of Voronoi-cell extents or of Voronoi-site positions
            (see ``input_type``)
        samples_param_values : ndarray
            a 2D numpy array where each row contains the parameter values
            associated with each Voronoi discretization found
            in ``samples_voronoi_cell_extents`` at the same row index
        position_bins: int or np.ndarray, optional
            the position bins or their number, default to 100
        param_value_bins: int or np.ndarray, optional
            the parameter value bins or their number, default to 100
        input_type : str, {'nuclei', 'extents'}
            argument determining whether each entry of `voronoi_cells` should be
            interpreted as a Voronoi-site position (``'nuclei'``) or as the
            extent of the Voronoi cell (``'extents'``)

        Returns
        -------
        density : ndarray, shape(nx, ny)
            The bi-dimensional histogram of samples x and y. Values in x are 
            histogrammed along the first dimension and values in y are histogrammed 
            along the second dimension
        X : ndarray, shape(nx+1,)
            The bin edges along the first dimension.
        Y : ndarray, shape(ny+1,)
            The bin edges along the second dimension.

        Examples
        --------
        .. code-block:: python

            from bayesbay.discretization import Voronoi1D

            # define and run the Bayesian inversion
            ...

            # plot
            results = inversion.get_results()
            samples_voronoi_sites = results["my_voronoi.discretization"]
            samples_param_values = results["my_voronoi.my_param_value"]
            density, X, Y = Voronoi1D.get_tessellation_density(
                samples_voronoi_sites, samples_param_values
            )
        """
        if input_type not in ["nuclei", "extents"]:
            raise ValueError("`input_type` should either be 'nuclei' or 'extents'")
        if isinstance(position_bins, int):
            lb = 0
            if input_type == "nuclei":
                ub = max([np.max(nuclei) for nuclei in samples_voronoi_cells])
            else:
                ub = 0
                for cell_extents in samples_voronoi_cells:
                    ub = max(ub, np.max(np.cumsum(np.array(cell_extents))))
            interp_positions = np.linspace(lb, ub, position_bins)
        elif isinstance(position_bins, np.ndarray):
            interp_positions = position_bins
        else:
            raise TypeError("`position_bins` should either be int or np.ndarray")
        interp_param_values = Voronoi1D._interpolate_tessellations(
            samples_voronoi_cells,
            samples_param_values,
            interp_positions,
            input_type=input_type,
        )
        density, X, Y = np.histogram2d(np.tile(interp_positions, interp_param_values.shape[0]),
                                       interp_param_values.ravel(),
                                       bins=(len(interp_positions), param_value_bins),
                                       density=True)
        return density, X, Y

    @staticmethod
    def plot_tessellation_density(
        samples_voronoi_cells: np.ndarray,
        samples_param_values: np.ndarray,
        position_bins: Union[int, np.ndarray] = 100,
        param_value_bins: Union[int, np.ndarray] = 100,
        ax=None,
        colorbar=True,
        swap_xy_axes=True,
        input_type="nuclei",
        **kwargs,
    ):
        """plot a 2D density histogram of the Voronoi tessellation

        Parameters
        ----------
        samples_voronoi_cells : list
            either a list of Voronoi-cell extents or of Voronoi-site positions
            (see ``input_type``)
        samples_param_values : ndarray
            a 2D numpy array where each row contains the parameter values
            associated with each Voronoi discretization found
            in ``samples_voronoi_cell_extents`` at the same row index
        position_bins: int or np.ndarray, optional
            the position bins or their number, default to 100
        param_value_bins: int or np.ndarray, optional
            the parameter value bins or their number, default to 100
        ax : Axes, optional
            an optional Axes object to plot on
        swap_xy_axes : bool
            if True (default), the x axis is swapped with the y axis so as to display
            the parameter value associated with each Voronoi cell on the x axis
        input_type : str, {'nuclei', 'extents'}
            argument determining whether each entry of `voronoi_cells` should be
            interpreted as a Voronoi-site position (``'nuclei'``) or as the
            extent of the Voronoi cell (``'extents'``)
        kwargs : dict, optional
            Additional keyword arguments to pass to ax.hist2d

        Returns
        -------
        ax : Axes
            The Axes object containing the 2D histogram
        cbar: Colorbar
            The Colorbar object associated with the 2D histogram

        Examples
        --------
        .. code-block:: python

            from bayesbay.discretization import Voronoi1D

            # define and run the Bayesian inversion
            ...

            # plot
            results = inversion.get_results()
            samples_voronoi_sites = results["my_voronoi.discretization"]
            samples_param_values = results["my_voronoi.my_param_value"]
            ax = Voronoi1D.plot_tessellation_density(
                samples_voronoi_sites, samples_param_values
            )
        """
        density, X, Y = Voronoi1D.get_tessellation_density(samples_voronoi_cells,
                                                           samples_param_values,
                                                           position_bins,
                                                           param_value_bins,
                                                           input_type)                                                     
        if ax is None:
            _, ax = plt.subplots()
        if swap_xy_axes:
            X, Y = Y, X
            if not ax.get_xlabel():
                ax.set_xlabel("Parameter values")
        else:
            if not ax.get_ylabel():
                ax.set_ylabel("Parameter values")
        
        img = ax.pcolormesh(X, Y, density, **kwargs)
        cbar = plt.colorbar(img, ax=ax, aspect=35, pad=0.02)
        cbar.set_label("Probability density")
        if ax.get_ylim()[0] < ax.get_ylim()[1] and swap_xy_axes:
            ax.invert_yaxis()
        return ax, cbar

    @staticmethod
    def plot_interface_hist(
        samples_voronoi_cells: np.ndarray,
        bins=100,
        ax=None,
        swap_xy_axes=True,
        input_type="nuclei",
        lb_tessellation=None,
        **kwargs,
    ):
        """plot the 1D histogram of Voronoi-interface positions

        Parameters
        ----------
        samples_voronoi_cells : list
            either a list of Voronoi-cell extents or of Voronoi-site positions
            (see ``input_type``)
        bins : int, optional
            number of histogram bins, by default 100
        ax : matplotlib.axes.Axes, optional
            an optional user-provided ax, by default None
        input_type : str, {'nuclei', 'extents'}
            argument determining whether each entry of `voronoi_cells` should be
            interpreted as a Voronoi-site position (``'nuclei'``) or as the
            extent of the Voronoi cell (``'extents'``)
        swap_xy_axes : bool
            if True (default), the x axis is swapped with the y axis so as to display
            the parameter value associated with each Voronoi cell on the x axis
        lb_tessellation : Number
            the lower boundary of the 1D tessellation, used to calculate the 
            interface positions when `input_type` is `'extents'`. Ignored otherwise.
        kwargs : dict, optional
            additional keyword arguments to pass to ax.bar
            
        Returns
        -------
        matplotlib.axes.Axes
        """
        positions = []
        for voronoi_cells in samples_voronoi_cells:
            positions.extend(Voronoi1D.compute_interface_positions(voronoi_cells, 
                                                                   input_type,
                                                                   lb_tessellation))
        if ax is None:
            _, ax = plt.subplots()
        hist, edges = np.histogram(positions, bins=bins, density=True)
        if swap_xy_axes:
            ax.barh(edges[:-1], hist, height=np.diff(edges), align="edge", **kwargs)
            if ax.get_ylim()[0] < ax.get_ylim()[1]:
                ax.invert_yaxis()
            if not ax.get_xlabel():
                ax.set_xlabel("Probability density")
        else:
            ax.bar(edges[:-1], hist, width=np.diff(edges), align="edge", **kwargs)
            if not ax.get_ylabel():
                ax.set_ylabel("Probability density")
        return ax

    @staticmethod
    def plot_tessellation(
        voronoi_cells: list,
        param_values: list,
        ax=None,
        bounds=(0, None),
        swap_xy_axes=True,
        input_type="nuclei",
        **kwargs,
    ):
        """plot multiple 1D Earth models based on sampled parameters.

        Parameters
        ----------
        voronoi_cells : list
            either Voronoi-cell extents or Voronoi-site positions (see
            ``input_type``)
        param_values : ndarray
            parameter values associated with each Voronoi cell
        ax : Axes, optional
            an optional Axes object to plot on
        bounds : tuple, optional
            lower and upper boundaries within which the tessellation will be
            displayed. Default is (0, None). When the upper boundary is None
            (default), this is determined by the maximum value in `voronoi_cells`
        swap_xy_axes : bool
            if True (default), the x axis is swapped with the y axis so as to display
            the parameter value associated with each Voronoi cell on the x axis
        input_type : str, {'nuclei', 'extents'}
            argument determining whether each entry of `voronoi_cells` should be
            interpreted as a Voronoi-site position (``'nuclei'``) or as the
            extent of the Voronoi cell (``'extents'``)
        kwargs : dict, optional
            additional keyword arguments to pass to ax.step

        Returns
        -------
        ax : Axes
            The Axes object containing the plot
        """
        lb, ub = bounds
        interface_positions = Voronoi1D.compute_interface_positions(voronoi_cells,
                                                                    input_type,
                                                                    lb)
        if ub is not None:
            assert ub > interface_positions[-1], (
                "`bounds[1]` should be greater"
                " than the sum of Voronoi"
                f" cell extents (here, {interface_positions[-1]})"
            )
            end_position = ub
        else:
            end_position = interface_positions[-1] + np.max(np.abs(interface_positions))/2

        x = np.insert(np.append(interface_positions, end_position), 0, lb)
        y = np.insert(param_values, 0, param_values[0])
        if swap_xy_axes:
            x, y = y, x

        if ax is None:
            _, ax = plt.subplots()

        # Default plotting style for samples
        sample_style = {
            "linewidth": kwargs.pop("linewidth", kwargs.pop("lw", 0.5)),
            "alpha": kwargs.pop("alpha", 1),
            "color": kwargs.pop(
                "color", kwargs.pop("c", "blue")
            ),  # Fixed color for the sample lines
        }
        sample_style.update(kwargs)  # Override with any provided kwargs
        ax.step(x, y, where="post", **sample_style)
        if ax.get_ylim()[0] < ax.get_ylim()[1] and swap_xy_axes:
            ax.invert_yaxis()
        if swap_xy_axes:
            if not ax.get_xlabel():
                ax.set_xlabel("Parameter values")
        else:
            if not ax.get_ylabel():
                ax.set_ylabel("Parameter values")
        return ax

    @staticmethod
    def plot_tessellations(
        samples_voronoi_cells: list,
        samples_param_values: list,
        ax=None,
        bounds=(0, None),
        swap_xy_axes=True,
        input_type="nuclei",
        **kwargs,
    ):
        """plot multiple 1D Earth models based on sampled parameters.

        Parameters
        ----------
        samples_voronoi_cells : list
            either a list of Voronoi-cell extents or of Voronoi-site positions
            (see ``input_type``)
        samples_param_values : ndarray
            a 2D numpy array where each row contains the parameter values
            associated with each Voronoi discretization found
            in ``samples_voronoi_cell_extents`` at the same row index
        ax : Axes, optional
            an optional Axes object to plot on
        bounds : tuple, optional
            lower and upper boundaries within which the tessellation will be
            displayed. Default is (0, None). When the upper boundary is None
            (default), this is determined by the maximum value in `voronoi_cells`
        swap_xy_axes : bool
            if True (default), the x axis is swapped with the y axis so as to display
            the parameter value associated with each Voronoi cell on the x axis
        input_type : str, {'nuclei', 'extents'}
            argument determining whether each entry of `voronoi_cells` should be
            interpreted as a Voronoi-site position (``'nuclei'``) or as the
            extent of the Voronoi cell (``'extents'``)
        kwargs : dict, optional
            additional keyword arguments to pass to ax.step

        Returns
        -------
        ax : Axes
            The Axes object containing the plot
        """
        lb, ub = bounds
        if input_type == "nuclei":
            samples_voronoi_cell_extents = [
                Voronoi1D.compute_cell_extents(nuclei, lb=lb)
                for nuclei in samples_voronoi_cells
            ]
        elif input_type == "extents":
            samples_voronoi_cell_extents = samples_voronoi_cells
        else:
            raise ValueError("`input_type` should either be 'nuclei' or 'extents'")

        if ax is None:
            _, ax = plt.subplots()
        if ub is not None:
            ax.set_ylim(0, ub)
        # Default plotting style for samples
        sample_style = {
            "linewidth": kwargs.pop("linewidth", kwargs.pop("lw", 0.5)),
            "alpha": kwargs.pop("alpha", 0.2),
            "color": kwargs.pop(
                "color", kwargs.pop("c", "blue")
            ),  # Fixed color for the sample lines
        }
        sample_style.update(kwargs)  # Override with any provided kwargs

        for extents, values in zip(samples_voronoi_cell_extents, samples_param_values):
            Voronoi1D.plot_tessellation(
                extents,
                values,
                **sample_style,
                ax=ax,
                bounds=bounds,
                input_type="extents",
            )

        if ax.get_ylim()[0] < ax.get_ylim()[1] and swap_xy_axes:
            ax.invert_yaxis()
        if swap_xy_axes:
            if not ax.get_xlabel():
                ax.set_xlabel("Parameter values")
        else:
            if not ax.get_ylabel():
                ax.set_ylabel("Parameter values")
        return ax

    @staticmethod
    def plot_tessellation_statistics(
        samples_voronoi_cells: list,
        samples_param_values: list,
        interp_positions: np.ndarray,
        percentiles=(10, 90),
        ax=None,
        input_type: str = "nuclei",
        swap_xy_axes: bool = True,
    ):
        """plot the mean, median, std and percentiles from the given samples

        Parameters
        ----------
        samples_voronoi_cells : list
            either a list of Voronoi-cell extents or of Voronoi-site positions
            (see ``input_type``)
        samples_param_values : list
            a list of parameter values to draw statistics from
        interp_positions : _type_
            points to interpolate
        percentiles : tuple, optional
            percentiles to calculate, by default (10, 90)
        ax : matplotlib.axes.Axes, optional
            an optional user-provided ax, by default None
        input_type : str, {'nuclei', 'extents'}
            argument determining whether each entry of `voronoi_cells` should be
            interpreted as a Voronoi-site position (``'nuclei'``) or as the
            extent of the Voronoi cell (``'extents'``)
        swap_xy_axes : bool
            if True (default), the x axis is swapped with the y axis so as to display
            the parameter value associated with each Voronoi cell on the x axis
        Returns
        -------
        matplotlib.axes.Axes
            the resulting plot that has the statistics on it
        """
        statistics = Voronoi1D.get_tessellation_statistics(
            samples_voronoi_cells,
            samples_param_values,
            interp_positions,
            percentiles,
            input_type=input_type,
        )

        if ax is None:
            _, ax = plt.subplots()
        _plot(
            interp_positions,
            statistics["mean"],
            ax,
            swap_xy_axes=swap_xy_axes,
            color="b",
            label="Mean",
        )
        _plot(
            interp_positions,
            statistics["mean"] - statistics["std"],
            ax,
            swap_xy_axes=swap_xy_axes,
            color="b",
            ls="--",
            label="STD",
        )
        _plot(
            interp_positions,
            statistics["mean"] + statistics["std"],
            ax,
            swap_xy_axes=swap_xy_axes,
            color="b",
            ls="--",
        )
        _plot(
            interp_positions,
            statistics["median"],
            ax,
            swap_xy_axes=swap_xy_axes,
            color="orange",
            label="Median",
        )
        _plot(
            interp_positions,
            statistics["percentiles"][0],
            ax,
            swap_xy_axes=swap_xy_axes,
            color="orange",
            ls="--",
            label=f"{percentiles[0]}-{percentiles[1]}th Perc.",
        )
        _plot(
            interp_positions,
            statistics["percentiles"][1],
            ax,
            swap_xy_axes=swap_xy_axes,
            color="orange",
            ls="--",
        )
        if ax.get_ylim()[0] < ax.get_ylim()[1] and swap_xy_axes:
            ax.invert_yaxis()
        ax.legend()
        return ax


class Voronoi2D(Voronoi):
    r"""Utility class for Voronoi tessellation in 1D

    Parameters
    ----------
    name : str
        name attributed to the Voronoi tessellation, for display and storing
        purposes
    vmin, vmax : Union[Number, np.ndarray]
        minimum/maximum value bounding each dimension. Ignored when
        ``polygon`` is not ``None``
    polygon: Union[np.ndarray, shapely.geometry.Polygon], optional
        polygon defining the domain of the Voronoi tessellation; Voronoi sites
        outside this polygon are not allowed
    perturb_std : Union[Number, np.ndarray]
        standard deviation of the Gaussians used to randomly perturb the Voronoi
        sites in each dimension.
    n_dimensions : Number, optional
        number of dimensions. None (default) results in a trans-dimensional
        discretization, with the dimensionality of the parameter space allowed
        to vary in the range ``n_dimensions_min``-``n_dimensions_max``
    n_dimensions_min, n_dimensions_max : Number, optional
        minimum and maximum number of dimensions, by default 1 and 10. These
        parameters are ignored if ``n_dimensions`` is not None, i.e. if the
        discretization is not trans-dimensional
    n_dimensions_init_range : Number, optional
        percentage of the range ``n_dimensions_min`` - ``n_dimensions_max`` used to
        initialize the number of dimensions (0.3. by default). For example, if
        ``n_dimensions_min`` = 1, ``n_dimensions_max`` = 10, and
        ``n_dimensions_init_range`` = 0.5,
        the maximum number of dimensions at the initialization is::

            int((n_dimensions_max - n_dimensions_min) * n_dimensions_init_range + n_dimensions_max)

    parameters : List[Parameter], optional
        a list of free parameters, by default None
    birth_from : {"prior", "neighbour"}, optional
        whether to initialize the free parameters associated with the newborn
        Voronoi cell by randomly drawing from their prior or by perturbing the
        value found in the nearest Voronoi cell (default)
    compute_kdtree : bool
        whether to compute a kd-tree for quick nearest-neighbor lookup at every
        perturbation of the discretization
    """

    def __init__(
        self,
        name: str,
        vmin: Number = None,
        vmax: Number = None,
        polygon: Union[np.ndarray, shapely.geometry.Polygon] = None,
        perturb_std: Union[Number, np.ndarray] = 1,
        n_dimensions: int = None,
        n_dimensions_min: int = 2,
        n_dimensions_max: int = 100,
        n_dimensions_init_range: Number = 0.3,
        parameters: List[Prior] = None,
        birth_from: str = "neighbour",  # either "neighbour" or "prior"
        compute_kdtree: bool = False,
    ):
        assert (vmin is not None and vmax is not None) or polygon is not None, (
            "Either `vmin`/`vmax` or `polygon`"
            " must not be None to properly define the discretization domain."
        )
        if polygon is not None:
            polygon = shapely.geometry.Polygon(polygon)
            vmin = polygon.bounds[:2]
            vmax = polygon.bounds[2:]
        self.polygon = polygon
        super().__init__(
            name=name,
            spatial_dimensions=2,
            vmin=vmin,
            vmax=vmax,
            perturb_std=perturb_std,
            n_dimensions=n_dimensions,
            n_dimensions_min=n_dimensions_min,
            n_dimensions_max=n_dimensions_max,
            n_dimensions_init_range=n_dimensions_init_range,
            parameters=parameters,
            birth_from=birth_from,
        )
        self.compute_kdtree = compute_kdtree

    def sample_site(self) -> np.ndarray:
        if self.polygon is not None:
            while True:
                new_site = super().sample_site()
                point = shapely.geometry.Point(new_site)
                if self.polygon.contains(point):
                    return new_site
        return super().sample_site()

    def _initialize(self) -> ParameterSpaceState:
        ps_state = super()._initialize()
        if self.compute_kdtree:
            return self._add_kdtree_to_ps_state(ps_state)
        return ps_state

    def _add_kdtree_to_ps_state(self, ps_state: ParameterSpaceState):
        voronoi_sites = ps_state.get_param_values("discretization")
        kdtree = scipy.spatial.KDTree(voronoi_sites)
        ps_state.save_to_cache("kdtree", kdtree)
        return ps_state

    def _perturb_site(
        self, site: Union[Number, np.ndarray]
    ) -> Union[Number, np.ndarray]:
        """perturbes a Voronoi  site

        Parameters
        ----------
        site : Union[Number, np.ndarray]
            Voronoi site position.

        Returns
        -------
        Union[Number, np.ndarray]
            perturbed Voronoi site position
        """
        if self.polygon is None:
            return super()._perturb_site(site)
        while True:
            random_deviate = np.random.normal(
                0, self.perturb_std, self.spatial_dimensions
            )
            new_site = site + random_deviate
            point = shapely.geometry.Point(new_site)
            if self.polygon.contains(point):
                return new_site

    def perturb_value(self, old_ps_state: ParameterSpaceState, isite: int):
        new_ps_state, log_prior_ratio = super().perturb_value(old_ps_state, isite)
        if self.compute_kdtree:
            new_ps_state = self._add_kdtree_to_ps_state(new_ps_state)
        return new_ps_state, log_prior_ratio

    def birth(
        self, old_ps_state: ParameterSpaceState
    ) -> Tuple[ParameterSpaceState, float]:
        new_ps_state, log_prob_ratio_birth = super().birth(old_ps_state)
        if self.compute_kdtree:
            new_ps_state = self._add_kdtree_to_ps_state(new_ps_state)
        return new_ps_state, log_prob_ratio_birth

    def death(self, old_ps_state: ParameterSpaceState):
        new_ps_state, log_prob_ratio_death = super().death(old_ps_state)
        if self.compute_kdtree:
            new_ps_state = self._add_kdtree_to_ps_state(new_ps_state)
        return new_ps_state, log_prob_ratio_death

    @staticmethod
    def interpolate_tessellation(
        voronoi_sites: np.ndarray,
        param_values: np.ndarray,
        interp_positions: np.ndarray,
    ):
        r"""nearest neighbour interpolation based on Voronoi-site
        positions and values associated with them.

        Parameters
        ----------
        voronoi_sites : (n, 2) np.ndarray
            the positions of the Voronoi sites
        param_values : (n,) np.ndarray
            the parameter values associated with each Voronoi cell
        query_points : (m, 2) np.ndarray
            the positions where interpolation is performed

        Returns
        -------
        np.ndarray
            interpolated values
        """
        kdtree = scipy.spatial.KDTree(voronoi_sites)
        inearest = kdtree.query(interp_positions)[1]
        return param_values[inearest]

    @staticmethod
    def _interpolate_tessellations(
        samples_voronoi_sites, samples_param_values, interp_positions
    ):
        interp_params = np.zeros((len(samples_param_values), len(interp_positions)))
        for i, (sample_sites, sample_values) in enumerate(
            zip(samples_voronoi_sites, samples_param_values)
        ):
            interp_params[i, :] = Voronoi2D.interpolate_tessellation(
                np.array(sample_sites), np.array(sample_values), interp_positions
            )
        return interp_params

    @staticmethod
    def get_tessellation_statistics(
        samples_voronoi_cells: list,
        samples_param_values: list,
        interp_positions: np.ndarray,
        percentiles: tuple = (10, 90),
    ) -> dict:
        """get the mean, median, std and percentiles of the given ensemble

        Parameters
        ----------
        samples_voronoi_cells : list
            either a list of Voronoi-cell extents or of Voronoi-site positions
            (see ``input_type``)
        samples_param_values : list
            a list of parameter values to draw statistics from
        interp_positions : np.ndarray
            points to interpolate
        percentiles : tuple, optional
            percentiles to calculate, by default (10, 90)

        Returns
        -------
        dict
            a dictionary with these keys: "mean", "median", "std" and "percentile"
        """
        interp_params = Voronoi2D._interpolate_tessellations(
            samples_voronoi_cells, samples_param_values, interp_positions
        )
        statistics = {
            "mean": np.mean(interp_params, axis=0),
            "median": np.median(interp_params, axis=0),
            "std": np.std(interp_params, axis=0),
            "percentiles": np.percentile(interp_params, percentiles, axis=0),
        }
        return statistics

    @staticmethod
    def plot_tessellation(
        voronoi_sites: np.ndarray,
        param_values: np.ndarray = None,
        ax=None,
        cmap="viridis",
        norm=None,
        vmin=None,
        vmax=None,
        voronoi_sites_kwargs=None,
        voronoi_plot_2d_kwargs=None,
    ):
        """display the Voronoi tessellation

        Parameters
        ----------
        voronoi_sites : np.ndarray of shape (m, 2)
            2D Voronoi-site positions
        param_values: np.ndarray, optional
            parameter values associated with each Voronoi cell. These could
            represent the physical property inferred in each cell of the
            discretized medium
        ax : matplotlib.axes.Axes, optional
            an optional Axes object to plot on
        cmap : Union[str, matplotlib.colors.Colormap]
            the Colormap instance or registered colormap name used to map scalar
            data to colors
        norm : Union[str, matplotlib.colors.Normalize]
            the normalization method used to scale scalar data to the [0, 1]
            range before mapping to colors using ``cmap``. By default, a linear
            scaling is used, mapping the lowest value to 0 and the highest to 1.
        vmin, vmax : Number
            minimum and maximum values used to create the colormap
        voronoi_sites_kwargs : dict
            keyword arguments passed to ``matplotlib.pyplot.plot``, used to
            plot the voronoi nuclei
        voronoi_plot_2d_kwargs : dict
            keyword arguments passed to ``scipy.spatial.voronoi_plot_2d``, used to
            plot the Voronoi interfaces
        Returns
        -------
        ax : matplotlib.axes.Axes
            The Axes object containing the 2D histogram
        """
        voronoi_plot_2d_kwargs = (
            voronoi_plot_2d_kwargs if voronoi_plot_2d_kwargs is not None else {}
        )
        interfaces_style = {
            "line_colors": "k",
            "show_vertices": False,
            "show_points": False,
            "line_width": 1,
        }
        interfaces_style.update(voronoi_plot_2d_kwargs)
        voronoi_sites_kwargs = (
            voronoi_sites_kwargs if voronoi_sites_kwargs is not None else {}
        )
        sites_style = {
            "color": voronoi_sites_kwargs.pop(
                "color", voronoi_sites_kwargs.pop("c", "k")
            ),
            "marker": voronoi_sites_kwargs.pop("marker", "o"),
            "ms": voronoi_sites_kwargs.pop(
                "ms", voronoi_sites_kwargs.pop("markersize", 2)
            ),
            "ls": "",
            "lw": 0,
        }
        sites_style.update(voronoi_sites_kwargs)

        xmax = np.max(np.abs(voronoi_sites[:, 0]))
        ymax = np.max(np.abs(voronoi_sites[:, 1]))
        sites = np.append(
            voronoi_sites,
            [
                [xmax * 100, ymax * 100],
                [-xmax * 100, ymax * 100],
                [xmax * 100, -ymax * 100],
                [-xmax * 100, -ymax * 100],
            ],
            axis=0,
        )

        voronoi = scipy.spatial.Voronoi(sites)
        if ax is None:
            fig, ax = plt.subplots()
        if param_values is not None:
            # make sure scipy.spatial.Voronoi didn't resort the original sites
            isort = [
                np.flatnonzero(np.all(p == voronoi.points, axis=1)).item()
                for i, p in enumerate(sites[:-4])
            ]
            ax, cbar = Voronoi2D._fill_tessellation(
                voronoi,
                param_values[isort],
                ax=ax,
                vmin=vmin,
                vmax=vmax,
                norm=norm,
                cmap=cmap,
            )
        else:
            cbar = None

        scipy.spatial.voronoi_plot_2d(voronoi, ax=ax, **interfaces_style)
        ax.plot(voronoi_sites[:, 0], voronoi_sites[:, 1], **sites_style)
        ax.set_xlabel("x")
        ax.set_ylabel("y")
        ax.set_xlim(voronoi_sites[:, 0].min(), voronoi_sites[:, 0].max())
        ax.set_ylim(voronoi_sites[:, 1].min(), voronoi_sites[:, 1].max())
        return ax, cbar

    @staticmethod
    def _fill_tessellation(
        voronoi: scipy.spatial.Voronoi,
        param_values: np.ndarray,
        ax=None,
        vmin=None,
        vmax=None,
        norm=None,
        cmap=None,
    ):
        if ax is None:
            fig, ax = plt.subplots()
        vmin = vmin if vmin is not None else min(param_values)
        vmax = vmax if vmax is not None else max(param_values)
        norm = norm if norm is not None else plt.Normalize(vmin=vmin, vmax=vmax)
        cmap = cmap if isinstance(cmap, mpl.colors.Colormap) else mpl.colormaps[cmap]
        colors = cmap(norm(param_values))

        for ipoint, iregion in enumerate(voronoi.point_region):
            region = voronoi.regions[iregion]
            if region and not -1 in region:  # Filter out points at infinity
                polygon = [voronoi.vertices[i] for i in region if i >= 0]
                ax.fill(*zip(*polygon), color=colors[ipoint])

        # Create a colorbar to show the mapping between values and colors
        cbar = ax.figure.colorbar(
            mpl.cm.ScalarMappable(cmap=cmap, norm=norm), ax=ax, aspect=35, pad=0.02
        )
        cbar.set_label("Parameter Values")
        return ax, cbar



from abc import abstractmethod, ABC
from typing import Tuple
from numbers import Number
from .._state import State, ParameterSpaceState


class Perturbation(ABC):
    """Base class for a perturbation callable

    Its subclasses represent different types of perturbations. The core functions of
    each subclass are :meth:`perturb` and :meth:`log_prior_ratio` and are to be called
    by :meth:`bayesbay.BaseMarkovChain.advance_chain` on each iteration.

    Users don't instantiate these perturbation classes directly; the instances are
    generated by :class:`bayesbay.parameterization.Parameterization` and
    :class:`bayesbay.LogLikelihood` instead.
    """

    @abstractmethod
    def perturb(self, state: State) -> Tuple[State, Number]:
        r"""proposes a new state from the given state and calculates the log of
        the corresponding partial acceptance probability

        .. math::
            \underbrace{\alpha_{p}}_{\begin{array}{c} \text{Partial} \\ \text{acceptance} \\ \text{probability} \end{array}} =
            \underbrace{\frac{p\left({\bf m'}\right)}{p\left({\bf m}\right)}}_{\text{Prior ratio}}
            \underbrace{\frac{q\left({\bf m} \mid {\bf m'}\right)}{q\left({\bf m'} \mid {\bf m}\right)}}_{\text{Proposal ratio}}
            \underbrace{\lvert \mathbf{J} \rvert}_{\begin{array}{c} \text{Jacobian} \\ \text{determinant} \end{array}}

        where :math:`{\bf d}_{obs}` denotes the observed data and
        :math:`\mathbf{J}` the Jacobian of the transformation.

        Parameters
        ----------
        state : State
            the given current state

        Returns
        -------
        Tuple[State, Number]
            the proposed new state and
            :math:`\alpha_{p} = \log(
            \frac{p({\bf m'})}{p({\bf m})}
            \frac{q\left({\bf m}
            \mid {\bf m'}\right)}{q\left({\bf m'} \mid {\bf m}\right)}
            \lvert \mathbf{J} \rvert)`
        """
        raise NotImplementedError

    def __call__(self, state: State) -> Tuple[State, Number]:
        """same as :meth:`perturb`

        Parameters
        ----------
        state : State
            the given current state

        Returns
        -------
        Tuple[State, Number]
            proposed new state and the partial acceptance probability excluding log
            likelihood ratio for this perturbation
        """
        return self.perturb(state)

    @property
    def type(self) -> str:
        """type of current perturbation (i.e. class name)

        Returns
        -------
        str
            the name of this current class
        """
        return self.__class__.__name__

    @property
    def __name__(self) -> str:
        """same as :attr:`type`

        Returns
        -------
        str
            the name of this current class
        """
        return self.type

    def __repr__(self) -> str:
        return self.__name__


class ParamSpaceMixin:
    @abstractmethod
    def perturb_param_space_state(
        self, ps_state: State
    ) -> Tuple[ParameterSpaceState, Number]:
        r"""proposes a new parameter space state from the given state and calculates
        the log of the corresponding partial acceptance probability

        Parameters
        ----------
        state : ParameterSpaceState
            the given current parameter space state

        Returns
        -------
        Tuple[State, Number]
            the proposed new parameter space state and the partial acceptance
            probability
        """
        raise NotImplementedError



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Dec 21 15:56:00 2022

@author: fabrizio
"""

from abc import ABC, abstractmethod
from typing import Any, Callable, Tuple, Union, Dict
from numbers import Number
import random
from functools import partial
import math
import numpy as np
import scipy

from ..exceptions._exceptions import OutOfDomainException
from .._utils_1d import interpolate_linear_1d


TWO_PI = 2 * math.pi
SQRT_TWO_PI = math.sqrt(TWO_PI)


class Prior(ABC):
    """Base class for defining the prior probability associated with a free parameter
    in the inference problem
    """

    def __init__(self, **kwargs):
        self._name = kwargs["name"]
        if "position" in kwargs and kwargs["position"] is not None:
            self.position = np.array(kwargs["position"], dtype=float)
        else:
            self.position = None
        assert "perturb_std" in kwargs
        self._repr_args = kwargs

    def __repr__(self) -> str:
        string = f"{self.__class__.__name__}("
        for k, v in self._repr_args.items():
            string += f"{k}={v}, "
        return f"{string[:-2]})"

    @property
    def name(self) -> str:
        return self._name

    @abstractmethod
    def sample(self, position: Number = None) -> Number:
        r"""sample a new value from the prior

        Paramters
        ---------
        position : Union[np.ndarray, Number], optional
            the position (in the discretization domain) associated with the value,
            None by default

        Returns
        -------
        Number
            a value corresponding to the prior probability (at the given position)
        """
        raise NotImplementedError

    @abstractmethod
    def initialize(self, positions: np.ndarray = None) -> np.ndarray:
        r"""initializes the values of this (possibly position-dependent)
        parameter

        This is the vectorized version of the method :meth:`sample`.

        Parameters
        ----------
        positions : np.ndarray, optional
            the position (in the discretization domain) associated with the value,
            None by default

        Returns
        -------
        np.ndarray
            an array of values or one value corresponding to the prior probability (at
            the given positions)
        """
        raise NotImplementedError

    @abstractmethod
    def perturb_value(
        self, value: Number, position: Number = None
    ) -> Tuple[Number, Number]:
        r"""perturbs the given value, in a way that may depend on the position
        associated with such a parameter value, and calculates the log of the
        corresponding partial acceptance probability

        .. math::
            \underbrace{\alpha_{p}}_{\begin{array}{c} \text{Partial} \\ \text{acceptance} \\ \text{probability} \end{array}} =
            \underbrace{\frac{p\left({\bf m'}\right)}{p\left({\bf m}\right)}}_{\text{Prior ratio}}
            \underbrace{\frac{q\left({\bf m} \mid {\bf m'}\right)}{q\left({\bf m'} \mid {\bf m}\right)}}_{\text{Proposal ratio}}
            \underbrace{\lvert \mathbf{J} \rvert}_{\begin{array}{c} \text{Jacobian} \\ \text{determinant} \end{array}},

        Parameters
        ----------
        value : Number
            the current value to be perturbed from
        position : Number, optional
            the position (in the discretization domain) associated with the value,
            None by default

        Returns
        -------
        Tuple[Number, Number]
            the perturbed value and
            :math:`\alpha_{p} = \log(
            \frac{p({\bf m'})}{p({\bf m})}
            \frac{q\left({\bf m}
            \mid {\bf m'}\right)}{q\left({\bf m'} \mid {\bf m}\right)}
            \lvert \mathbf{J} \rvert)`
        """
        raise NotImplementedError

    @abstractmethod
    def log_prior(self, value: Number, position: Number = None) -> Number:
        r"""calculates the log of the prior probability of occurrence of the
        given value, which may depend on the considered position

        Parameters
        ----------
        value : Number
            the value to calculate the probability density for
        position : Number, optional
            the position (in the discretization domain) associated with the value,
            None by default

        Returns
        -------
        Number
            the log prior probability density
        """
        raise NotImplementedError

    def set_custom_initialize(
        self,
        initialize_func: Callable[["Prior", np.ndarray], np.ndarray],
    ):
        r"""sets a custom initialization function

        Parameters
        ----------
        initialize_func: Callable[["Prior", np.ndarray], np.ndarray]
            The function to use for initialization. This function should take a
            :class:`Prior` instance and optionally an array of positions as input
            arguments, and produce an array of values as output.

        Examples
        --------
        .. code-block:: python

            def my_init(
                param: bb.prior.Prior,
                position: np.ndarray
            ) -> np.ndarray:
                print("This is my custom init!")
                return np.ones(len(position))

            my_param.set_custom_initialize(my_init)
        """
        if not callable(initialize_func):
            raise ValueError("initialize_func must be a callable function.")
        self.initialize = partial(self._initializer, initialize_func)

    def _initializer(self, initialize_func, *args, **kwargs):
        return initialize_func(self, *args, **kwargs)

    def get_perturb_std(self, position: Union[Number, np.ndarray] = None) -> Number:
        r"""get the standard deviation of the Gaussian used to perturb
        the parameter, which may be dependent on the position in the
        discretization domain

        Parameters
        ----------
        position: Union[Number, np.ndarray], optional
            the position in the discretization domain at which the standard
            deviation of the Gaussian used to perturb the parameter is returned.
            Default is None

        Returns
        -------
        Number
            standard deviation of the Gaussian used to perturb the parameter,
            possibly at the specified position
        """
        if self.has_hyper_param("perturb_std"):
            return self.get_hyper_param("perturb_std", position)
        else:
            raise NotImplementedError("`get_perturb_std` needs to be implemented")

    def get_perturb_std_birth(
        self, position: Union[Number, np.ndarray] = None
    ) -> Number:
        r"""get the standard deviation of the Gaussian used to perturb
        the parameter at birth, which may be dependent on the position in the
        discretization domain

        Parameters
        ----------
        position: Union[Number, np.ndarray], optional
            the position in the discretization domain at which the standard
            deviation of the Gaussian used to perturb the parameter at birth
            is returned. Default is None

        Returns
        -------
        Number
            standard deviation of the Gaussian used to perturb the parameter at birth,
            possibly at the specified position
        """
        if self.has_hyper_param("perturb_std_birth"):
            return self.get_hyper_param("perturb_std_birth", position)
        else:
            return self.get_perturb_std(position)

    def add_hyper_params(self, hyper_params: Dict[str, Union[Number, np.ndarray]]):
        r"""Sets the attributes from the given dict and checks for errors

        Parameters
        ----------
        hyper_params : Dict[str, Union[Number, np.ndarray]]
            dictionary of attributes to be set
        """
        msg_scalar_when_pos_none = "should be a scalar when `position` is `None`"
        msg_scalar_or_same_len = (
            "should either be a scalar or have the same length as `position`"
        )
        for name, param_val in hyper_params.items():
            if self.position is None:
                assert np.isscalar(param_val), f"`{name}` {msg_scalar_when_pos_none}"
            else:
                assert np.isscalar(param_val) or len(param_val) == len(
                    self.position
                ), f"`{name}` {msg_scalar_or_same_len}"
            if not np.isscalar(param_val):
                param_val = np.array(param_val, dtype=float)
            setattr(self, name, self._init_hyper_param(param_val))

    def has_hyper_param(self, hyper_param: str) -> bool:
        r"""Whether or not the :class:`Prior` instance has the specified
        attribute

        Parameters
        ----------
        hyper_param : str
            the name of the attribute
        """
        return hasattr(self, hyper_param)

    def get_hyper_param(
        self, hyper_param: str, position: Union[Number, np.ndarray] = None
    ) -> Union[Number, np.ndarray]:
        r"""Retrieves the value corresponding to the specified attribute, which
        may be a function of position

        Parameters
        ----------
        hyper_param : str
            the name of the attribute
        position : Union[np.ndarray, Number], optional
            the position (in the discretization domain) associated with the value,
            None by default

        Returns
        -------
        Union[Number, np.ndarray]
            value corresponding to the specified attribute
        """
        _hyper_param = getattr(self, hyper_param)
        if np.isscalar(_hyper_param):
            return _hyper_param
        elif position is not None:
            return _hyper_param(position)
        else:
            raise ValueError(
                f"`{hyper_param}` is dependent on position but got None for it"
            )

    def _init_hyper_param(self, hyper_param: Union[Number, np.ndarray]):
        # to be called after self.position is assigned
        if np.isscalar(hyper_param):
            return hyper_param
        elif np.ndim(self.position) == 1:
            return partial(interpolate_linear_1d, x=self.position, y=hyper_param)
        else:
            interpolator = scipy.interpolate.LinearNDInterpolator(
                points=self.position,
                values=hyper_param,
            )
            return _interpolate_linear_nd(self.name, interpolator)


class UniformPrior(Prior):
    r"""Class for defining the prior probability of a free parameter distributed
    according to a uniform probability distribution

    Parameters
    ----------
    name : str
        name of the parameter, for display and storing purposes
    vmin : Union[Number, np.ndarray]
        the lower bound for this parameter. This can either be a scalar or an array
        if the defined probability distribution is a function of ``position``
        in the discretization domain
    vmax : Union[Number, np.ndarray]
        the upper bound for this parameter. This can either be a scalar or an array
        if the defined probability distribution is a function of ``position``
        in the discretization domain
    perturb_std : Union[Number, np.ndarray]
        standard deviation of the Gaussians used to randomly perturb the parameter.
        This can either be a scalar or an array if the defined probability
        distribution is a function of ``position`` in the discretization domain
    position : np.ndarray, optional
        position in the discretization domain, used to define a position-dependent
        probability distribution. None by default
    """

    def __init__(
        self,
        name: str,
        vmin: Union[Number, np.ndarray],
        vmax: Union[Number, np.ndarray],
        perturb_std: Union[Number, np.ndarray],
        perturb_std_birth: Union[Number, np.ndarray] = None,
        position: np.ndarray = None,
    ):
        super().__init__(
            name=name,
            position=position,
            vmin=vmin,
            vmax=vmax,
            perturb_std=perturb_std,
            perturb_std_birth=perturb_std_birth,
        )
        self.add_hyper_params(
            {
                "vmin": vmin,
                "vmax": vmax,
                "perturb_std": perturb_std,
                "perturb_std_birth": perturb_std_birth
                if perturb_std_birth is not None
                else perturb_std,
                "delta": np.array(vmax) - np.array(vmin),
            }
        )

    def get_delta(self, position: Union[Number, np.ndarray] = None) -> Number:
        r"""get the range :math:`\Delta v = v_{max} - v_{min}`, which may be
        dependent on the specified position in the discretization domain

        Parameters
        ----------
        position: Union[Number, np.ndarray], optional
            the position in the discretization domain at which the uniform
            distribution range will be returned. None by default

        Returns
        -------
        Number
            the range :math:`\Delta v = v_{max} - v_{min}`
        """
        return self.get_hyper_param("delta", position)

    def get_vmin_vmax(
        self, position: Union[Number, np.ndarray] = None
    ) -> Tuple[Number, Number]:
        r"""get the lower and upper bounds of the parameter, which may be
        dependent on position in the discretization domain

        Parameters
        ----------
        position: Union[Number, np.ndarray], optional
            the position in the discretization domain at which the the lower
            and upper bounds of the parameter will be returned. None by default

        Returns
        -------
        Tuple[Number, Number]
            the lower (``vmin``) and upper (``vmax``)
        """
        vmin = self.get_hyper_param("vmin", position)
        vmax = self.get_hyper_param("vmax", position)
        return vmin, vmax

    def sample(self, position: Union[Number, np.ndarray] = None) -> Number:
        vmin, vmax = self.get_vmin_vmax(position)
        return random.uniform(vmin, vmax)

    def initialize(self, positions: np.ndarray = None) -> np.ndarray:
        vmin, vmax = np.array([self.get_vmin_vmax(p) for p in positions]).T
        return np.random.uniform(vmin, vmax, len(positions))

    def perturb_value(
        self,
        value: Number,
        position: Union[Number, np.ndarray] = None,
        is_birth: bool = False,
    ) -> Tuple[Number, Number]:
        r"""perturbs the given value, in a way that may depend on the position
        in the discretization domain and calculates the log of the corresponding
        partial acceptance probability,

        .. math::
            \underbrace{\alpha_{p}}_{\begin{array}{c} \text{Partial} \\ \text{acceptance} \\ \text{probability} \end{array}} =
            \underbrace{\frac{p\left({\bf m'}\right)}{p\left({\bf m}\right)}}_{\text{Prior ratio}}
            \underbrace{\frac{q\left({\bf m} \mid {\bf m'}\right)}{q\left({\bf m'} \mid {\bf m}\right)}}_{\text{Proposal ratio}}
            \underbrace{\lvert \mathbf{J} \rvert}_{\begin{array}{c} \text{Jacobian} \\ \text{determinant} \end{array}},

        which in this case equals zero. In the above equation, :math:`\bf m'` denotes
        the perturbed model as obtained through a random deviate from a normal
        distribution :math:`\mathcal{N}(v, \sigma)`, where :math:`v` denotes
        the original ``value`` and :math:`\sigma` the standard deviation of the
        Gaussian used for the perturbation. :math:`\sigma` may be dependent
        on the specified position (:attr:`UniformPrior.perturb_std`).

        Parameters
        ----------
        value : Number
            the current value to be perturbed from
        position : Union[Number, np.ndarray], optional
            the position in the discretization domain at which the parameter
            ``value`` is to be perturbed
        is_birth : bool, optional
            whether the perturbation is a birth or not, False by default

        Returns
        -------
        Tuple[Number, Number]
            the perturbed value and
            :math:`\alpha_{p} = \log(
            \frac{p({\bf m'})}{p({\bf m})}
            \frac{q\left({\bf m}
            \mid {\bf m'}\right)}{q\left({\bf m'} \mid {\bf m}\right)}
            \lvert \mathbf{J} \rvert) = 0`
        """
        # randomly perturb the value until within range
        if is_birth:
            std = self.get_perturb_std_birth(position)
        else:
            std = self.get_perturb_std(position)
        vmin, vmax = self.get_vmin_vmax(position)
        random_deviate = random.normalvariate(0, std)
        new_value = value + random_deviate
        return new_value, 0 if vmin <= new_value <= vmax else -math.inf

    def log_prior(
        self, value: Number, position: Union[Number, np.ndarray] = None
    ) -> Number:
        r"""calculates the log of the prior probability density for the given
        value, which may be dependent on the position in the discretization
        domain

        Parameters
        ----------
        value : Number
            the value to calculate the probability density for
        position : Union[Number, np.ndarray], optional
            the position in the discretization domain at which the prior
            probability of the parameter ``value`` is to be retrieved

        Returns
        -------
        Number
            the log of the prior probability :math:`p(v) = \frac{1}{\Delta v}`,
            where :math:`v` denotes the ``value`` passed by the user and
            :math:`\Delta v = v_{max} - v_{min}` the range within which the
            uniform parameter is allowed to vary. This may be
            dependent on position in the discretization domain
        """
        vmin, vmax = self.get_vmin_vmax(position)
        if vmin <= value <= vmax:
            return -math.log(vmax - vmin)
        else:
            return -math.inf


class GaussianPrior(Prior):
    r"""Class for defining the prior probability of a free parameter distributed
    according to a Gaussian distribution, :math:`\mathcal{N}(\mu, \sigma)`,
    where :math:`\mu` denotes the mean and :math:`\sigma` the standard deviation

    Parameters
    ----------
    name : str
        name of the parameter, for display and storing purposes
    mean : Union[Number, np.ndarray]
        mean of the Gaussian. This can either be a scalar or an array
        if the defined probability distribution is a function of ``position``
        in the discretization domain
    std : Union[Number, np.ndarray]
        standard deviation of the Gaussian. This can either be a scalar or an array
        if the defined probability distribution is a function of ``position``
        in the discretization domain
    perturb_std : Union[Number, np.ndarray]
        standard deviation of the Gaussians used to randomly perturb the parameter.
        This can either be a scalar or an array if the defined probability distribution
        is a function of ``position`` in the discretization domain
    position : np.ndarray, optional
        position in the discretization domain, used to define a position-dependent
        probability distribution. None by default
    """

    def __init__(
        self,
        name: str,
        mean: Union[Number, np.ndarray],
        std: Union[Number, np.ndarray],
        perturb_std: Union[Number, np.ndarray],
        perturb_std_birth: Union[Number, np.ndarray] = None,
        position: np.ndarray = None,
    ):
        super().__init__(
            name=name,
            position=position,
            mean=mean,
            std=std,
            perturb_std=perturb_std,
            perturb_std_birth=perturb_std_birth,
        )
        self.add_hyper_params(
            {
                "mean": mean,
                "std": std,
                "perturb_std": perturb_std,
                "perturb_std_birth": perturb_std_birth
                if perturb_std_birth is not None
                else perturb_std,
            }
        )

    def get_mean(self, position: Union[Number, np.ndarray] = None) -> Number:
        """get the mean of the Gaussian parameter, which may be dependent on
        position in the discretization domain

        Parameters
        ----------
        position: Union[Number, np.ndarray], optional
            position in the discretization domain at which the mean of the
            Gaussian will be returned. None by default

        Returns
        -------
        Number
            mean at the given position
        """
        return self.get_hyper_param("mean", position)

    def get_std(self, position: Union[Number, np.ndarray] = None) -> Number:
        """get the standard deviation of the Gaussian, which may be
        dependent on position in the discretization domain

        Parameters
        ----------
        position: Union[Number, np.ndarray], optional
            position in the discretization domain at which the standard
            deviation of the Gaussian will be returned. None by default

        Returns
        -------
        Number
            standard deviation at the given position
        """
        return self.get_hyper_param("std", position)

    def sample(self, position: Number = None) -> Number:
        mean = self.get_mean(position)
        std = self.get_std(position)
        return random.gauss(mean, std)

    def initialize(self, positions: np.ndarray = None) -> np.ndarray:
        r"""initialize the parameter, possibly at specific positions in the
        discretization domain

        Parameters
        ----------
        positions: np.ndarray, optional
            the positions in the discretization domain at which the parameter is
            initialized. None by default

        Returns
        -------
        np.ndarray
            an array of values or one value corresponding to the given positions,
            chosen according to the normal distribution defined
            by :attr:`mean` and :attr:`std` at the given positions
        """
        mean = self.get_mean(positions)
        std = self.get_std(positions)
        values = np.random.normal(mean, std, len(positions))
        return values

    def perturb_value(
        self,
        value: Number,
        position: Union[Number, np.ndarray] = None,
        is_birth: bool = False,
    ) -> Tuple[Number, Number]:
        r"""perturbs the given value, in a way that may depend on the position
        in the discretization, and calculates the log of the corresponding
        partial acceptance probability,

        .. math::
            \underbrace{\alpha_{p}}_{\begin{array}{c} \text{Partial} \\ \text{acceptance} \\ \text{probability} \end{array}} =
            \underbrace{\frac{p\left({\bf m'}\right)}{p\left({\bf m}\right)}}_{\text{Prior ratio}}
            \underbrace{\frac{q\left({\bf m} \mid {\bf m'}\right)}{q\left({\bf m'} \mid {\bf m}\right)}}_{\text{Proposal ratio}}
            \underbrace{\lvert \mathbf{J} \rvert}_{\begin{array}{c} \text{Jacobian} \\ \text{determinant} \end{array}},

        where :math:`\bf m'` denotes the perturbed model as obtained through a
        random deviate from a normal distribution :math:`\mathcal{N}(v, \theta)`,
        where :math:`v` denotes the original ``value`` and :math:`\theta` the
        standard deviation of the Gaussian used for the perturbation.
        :math:`\theta` may be dependent on the specified position
        (:attr:`GaussianPrior.perturb_std`).

        Parameters
        ----------
        value : Number
            the current value to be perturbed from
        position : Number
            the position of the value to be perturbed
        is_birth : bool, optional
            whether the perturbation is a birth or not, False by default

        Returns
        -------
        Tuple[Number, Number]
            the perturbed value and
            :math:`\alpha_{p} = \log(
            \frac{p({\bf m'})}{p({\bf m})}
            \frac{q\left({\bf m}
            \mid {\bf m'}\right)}{q\left({\bf m'} \mid {\bf m}\right)}
            \lvert \mathbf{J} \rvert)`
        """
        if is_birth:
            perturb_std = self.get_perturb_std_birth(position)
        else:
            perturb_std = self.get_perturb_std(position)
        random_deviate = random.normalvariate(0, perturb_std)
        new_value = value + random_deviate
        mean = self.get_mean(position)
        std = self.get_std(position)
        ratio = (value - new_value) * (value + new_value - 2 * mean) / (2 * std**2)
        return new_value, ratio

    def log_prior(
        self, value: Number, position: Union[Number, np.ndarray] = None
    ) -> Number:
        r"""calculates the log of the prior probability density for the given
        value, which may be dependent on the position in the discretization
        domain

        Parameters
        ----------
        value : Number
            the value to calculate the probability density for
        position : Union[Number, np.ndarray], optional
            the position in the discretization domain at which the prior
            probability of the parameter ``value`` is to be retrieved

        Returns
        -------
        Number
            the log of the prior probability :math:`p(v) = \frac{1}{\sigma \sqrt{2 \pi}}
            \exp \Big \lbrace -\frac{\left( v - \mu \right)^2}
            {2\sigma^2} \Big \rbrace`, where :math:`\mu` and :math:`\sigma`
            denote the (prior) mean and standard deviation of the
            Gaussian and :math:`v` the ``value`` at which the prior
            probability is to be retrieved. :math:`\mu` and :math:`\sigma` may
            be dependent on position in the discretization domain.
        """
        mean = self.get_mean(position)
        std = self.get_std(position)
        return (
            -0.5 * np.log(2 * np.pi) - np.log(std) - 0.5 * ((value - mean) / std) ** 2
        )


class LaplacePrior(Prior):
    r"""Class for defining the prior probability of a free parameter distributed
    according to a Laplace distribution, with probability density function
    :math:`p(v) = \frac{1}{2b} \exp \Big \lbrace -\frac{|v - \mu|}{b} \Big \rbrace`,
    where :math:`\mu` denotes the mean and :math:`b` the scale.

    Parameters
    ----------
    name : str
        name of the parameter, for display and storing purposes
    mean : Union[Number, np.ndarray]
        mean of the Gaussian. This can either be a scalar or an array
        if the defined probability distribution is a function of ``position``
        in the discretization domain
    std : Union[Number, np.ndarray]
        standard deviation of the Gaussian. This can either be a scalar or an array
        if the defined probability distribution is a function of ``position``
        in the discretization domain
    perturb_std : Union[Number, np.ndarray]
        standard deviation of the Gaussians used to randomly perturb the parameter.
        This can either be a scalar or an array if the defined probability distribution
        is a function of ``position`` in the discretization domain
    position : np.ndarray, optional
        position in the discretization domain, used to define a position-dependent
        probability distribution. None by default
    """

    def __init__(
        self,
        name: str,
        mean: Union[Number, np.ndarray],
        scale: Union[Number, np.ndarray],
        perturb_std: Union[Number, np.ndarray],
        perturb_std_birth: Union[Number, np.ndarray] = None,
        position: np.ndarray = None,
    ):
        super().__init__(
            name=name,
            position=position,
            mean=mean,
            scale=scale,
            perturb_std=perturb_std,
        )
        self.add_hyper_params(
            {
                "mean": mean,
                "scale": scale,
                "perturb_std": perturb_std,
                "perturb_std_birth": perturb_std_birth
                if perturb_std_birth is not None
                else perturb_std,
            }
        )

    def get_mean(self, position: Union[Number, np.ndarray] = None) -> Number:
        """get the mean of the Laplace parameter, which may be dependent on
        position in the discretization domain

        Parameters
        ----------
        position: Union[Number, np.ndarray], optional
            position in the discretization domain at which the mean of the
            Laplace distribution will be returned. None by default

        Returns
        -------
        Number
            mean at the given position
        """
        return self.get_hyper_param("mean", position)

    def get_scale(self, position: Union[Number, np.ndarray] = None) -> Number:
        """get the scale of the Laplace distribution, which may be
        dependent on position in the discretization domain

        Parameters
        ----------
        position: Union[Number, np.ndarray], optional
            position in the discretization domain at which the scale
            of the Laplace distribution will be returned. None by default

        Returns
        -------
        Number
            scale at the given position
        """
        return self.get_hyper_param("scale", position)

    def sample(self, position: Number = None) -> Number:
        mean = self.get_mean(position)
        scale = self.get_scale(position)
        return np.random.laplace(mean, scale)

    def initialize(self, positions: np.ndarray = None) -> np.ndarray:
        r"""initialize the parameter, possibly at specific positions in the
        discretization domain

        Parameters
        ----------
        positions: np.ndarray, optional
            the positions in the discretization domain at which the parameter is
            initialized. None by default

        Returns
        -------
        np.ndarray
            an array of values or one value corresponding to the given positions,
            chosen according to the Laplace distribution defined
            by :attr:`mean` and :attr:`scale` at the given positions
        """
        mean = self.get_mean(positions)
        scale = self.get_scale(positions)
        values = np.random.laplace(mean, scale, len(positions))
        return values

    def perturb_value(
        self,
        value: Number,
        position: Union[Number, np.ndarray] = None,
        is_birth: bool = False,
    ) -> Tuple[Number, Number]:
        r"""perturbs the given value, in a way that may depend on the position
        in the discretization, and calculates the log of the corresponding
        partial acceptance probability,

        .. math::
            \underbrace{\alpha_{p}}_{\begin{array}{c} \text{Partial} \\ \text{acceptance} \\ \text{probability} \end{array}} =
            \underbrace{\frac{p\left({\bf m'}\right)}{p\left({\bf m}\right)}}_{\text{Prior ratio}}
            \underbrace{\frac{q\left({\bf m} \mid {\bf m'}\right)}{q\left({\bf m'} \mid {\bf m}\right)}}_{\text{Proposal ratio}}
            \underbrace{\lvert \mathbf{J} \rvert}_{\begin{array}{c} \text{Jacobian} \\ \text{determinant} \end{array}},

        where :math:`\bf m'` denotes the perturbed model as obtained through a
        random deviate from a normal distribution :math:`\mathcal{N}(v, \theta)`,
        where :math:`v` denotes the original ``value`` and :math:`\theta` the
        standard deviation of the Gaussian used for the perturbation.
        :math:`\theta` may be dependent on the specified position
        (:attr:`LaplacePrior.perturb_std`).

        Parameters
        ----------
        value : Number
            the current value to be perturbed from
        position : Number
            the position of the value to be perturbed
        is_birth : bool, optional
            whether the perturbation is a birth or not, False by default

        Returns
        -------
        Tuple[Number, Number]
            the perturbed value and
            :math:`\alpha_{p} = \log(
            \frac{p({\bf m'})}{p({\bf m})}
            \frac{q\left({\bf m}
            \mid {\bf m'}\right)}{q\left({\bf m'} \mid {\bf m}\right)}
            \lvert \mathbf{J} \rvert)`
        """
        if is_birth:
            perturb_std = self.get_perturb_std_birth(position)
        else:
            perturb_std = self.get_perturb_std(position)
        random_deviate = random.normalvariate(0, perturb_std)
        new_value = value + random_deviate
        mean = self.get_mean(position)
        scale = self.get_scale(position)
        ratio = (abs(value - mean) - abs(new_value - mean)) / scale
        return new_value, ratio

    def log_prior(
        self, value: Number, position: Union[Number, np.ndarray] = None
    ) -> Number:
        r"""calculates the log of the prior probability density for the given
        value, which may be dependent on the position in the discretization
        domain

        Parameters
        ----------
        value : Number
            the value to calculate the probability density for
        position : Union[Number, np.ndarray], optional
            the position in the discretization domain at which the prior
            probability of the parameter ``value`` is to be retrieved

        Returns
        -------
        Number
            the log of the prior probability :math:`p(v) = \frac{1}{2b}
            \exp \Big \lbrace -\frac{|v - \mu|}
            {b} \Big \rbrace`, where :math:`\mu` and :math:`b`
            denote the mean and scale of the Laplace distribution
            and :math:`v` the ``value`` at which the prior
            probability is to be retrieved. :math:`\mu` and :math:`b` may
            be dependent on position in the discretization domain.
        """
        mean = self.get_mean(position)
        scale = self.get_scale(position)
        return -math.log(2) - math.log(scale) - abs(value - mean) / scale


class CustomPrior(Prior):
    """Class enabling the definition of an arbitrary prior for a free parameter

    Parameters
    ----------
    name : str
        name of the current parameter, for display and storing purposes
    log_prior : Callable[[Number, Number], Number]
    sample : Callable[[Number], Number]
        a function that samples a new value (optionally dependent on a position). The
        parameter initialization will also be done by calling this function
    perturb_std : Union[Number, np.ndarray]
        standard deviation of the Gaussians used to randomly perturb the parameter.
        This can either be a scalar or an array if the defined probability distribution
        is a function of ``position`` in the discretization domain
    position : np.ndarray, optional
        position in the discretization domain, used to define a position-dependent
        probability distribution. None by default
    """

    def __init__(
        self,
        name: str,
        log_prior: Callable[[Number, Number], Number],
        sample: Callable[[Number], Number],
        perturb_std: Union[Number, np.ndarray],
        perturb_std_birth: Union[Number, np.ndarray] = None,
        position: np.ndarray = None,
    ):
        super().__init__(
            name=name,
            log_prior=log_prior,
            sample=sample,
            perturb_std=perturb_std,
            position=position,
        )
        self.add_hyper_params(
            {
                "perturb_std": perturb_std,
                "perturb_std_birth": perturb_std_birth
                if perturb_std_birth is not None
                else perturb_std,
            }
        )
        self._log_prior = log_prior
        self._sample = sample

    def sample(self, position: Union[Number, np.ndarray] = None) -> Number:
        return self._sample(position)

    def initialize(self, positions: np.ndarray = None) -> np.ndarray:
        result = np.empty_like(positions)
        for i, position in enumerate(positions):
            result[i] = self._sample(position)
        return result

    def perturb_value(
        self,
        value: Number,
        position: Union[Number, np.ndarray] = None,
        is_birth: bool = False,
    ) -> Tuple[Number, Number]:
        r"""perturbs the given value, in a way that may depend on the position
        in the discretization domain, and calculates the log of the corresponding
        partial acceptance probability,

        .. math::
            \underbrace{\alpha_{p}}_{\begin{array}{c} \text{Partial} \\ \text{acceptance} \\ \text{probability} \end{array}} =
            \underbrace{\frac{p\left({\bf m'}\right)}{p\left({\bf m}\right)}}_{\text{Prior ratio}}
            \underbrace{\frac{q\left({\bf m} \mid {\bf m'}\right)}{q\left({\bf m'} \mid {\bf m}\right)}}_{\text{Proposal ratio}}
            \underbrace{\lvert \mathbf{J} \rvert}_{\begin{array}{c} \text{Jacobian} \\ \text{determinant} \end{array}},

        where :math:`\bf m'` denotes the perturbed model as obtained through a
        random deviate from a normal distribution :math:`\mathcal{N}(v, \theta)`,
        where :math:`v` denotes the original ``value`` and :math:`\theta` the
        standard deviation of the Gaussian used for the perturbation.
        :math:`\theta` may be dependent on the specified position in the
        discretization domain (:attr:`CustomPrior.perturb_std`).

        Parameters
        ----------
        value : Number
            the current value to be perturbed from
        position : Number, optional
            the position of the value to be perturbed
        is_birth : bool, optional
            whether the perturbation is a birth or not, False by default

        Returns
        -------
        Tuple[Number, Number]
            the perturbed value and
            :math:`\alpha_{p} = \log(
            \frac{p({\bf m'})}{p({\bf m})}
            \frac{q\left({\bf m}
            \mid {\bf m'}\right)}{q\left({\bf m'} \mid {\bf m}\right)}
            \lvert \mathbf{J} \rvert)`
        """
        if is_birth:
            perturb_std = self.get_perturb_std_birth(position)
        else:
            perturb_std = self.get_perturb_std(position)
        random_deviate = random.normalvariate(0, perturb_std)
        new_value = value + random_deviate
        ratio = self.log_prior(new_value, position) - self.log_prior(value, position)
        return new_value, ratio

    def log_prior(
        self, value: Number, position: Union[Number, np.ndarray] = None
    ) -> Number:
        r"""calculates the log of the prior probability density for the given
        value, which may be dependent on the position in the discretization
        domain

        Parameters
        ----------
        value : Number
            the value to calculate the probability density for
        position : Union[Number, np.ndarray], optional
            the position in the discretization domain at which the prior
            probability of the parameter ``value`` is to be retrieved

        Returns
        -------
        Number
            the log prior probability density
        """
        try:
            return self._log_prior(value, position)
        except:
            return self._log_prior(value)


def _interpolate_linear_nd(variable_name, interpolator):
    def func(x):
        y_interp = interpolator(x)
        if np.isnan(y_interp):
            raise OutOfDomainException(variable_name, x)
        return y_interp.item()

    return func



import warnings

from ._state import DataNoiseState, ParameterSpaceState, State
from .likelihood._target import Target as _Target
from .likelihood._log_likelihood import LogLikelihood as _LogLikelihood
from ._markov_chain import MarkovChain, BaseMarkovChain
from ._bayes_inversion import BayesianInversion, BaseBayesianInversion

from . import (
    samplers,
    prior,
    likelihood,
    perturbations,
    parameterization,
    discretization,
)

from ._version import __version__


def Target(*args, **kwargs):
    warnings.warn(
        (
            "The 'Target' class has been moved to the 'likelihood' module. Please use "
            "'from bayesbay.likelihood import Target' instead."
        ),
        DeprecationWarning,
        stacklevel=2,
    )
    return _Target(*args, **kwargs)


def LogLikelihood(*args, **kwargs):
    warnings.warn(
        (
            "The 'LogLikelihood' class has been moved to the 'likelihood' module. "
            "Please use 'from bayesbay.likelihood import LogLikelihood' instead."
        ),
        DeprecationWarning,
        stacklevel=2,
    )
    return _LogLikelihood(*args, **kwargs)


__all__ = [
    "DataNoiseState",
    "ParameterSpaceState",
    "State",
    "Target",
    "LogLikelihood",
    "MarkovChain",
    "BaseMarkovChain",
    "BayesianInversion",
    "BaseBayesianInversion",
]



# cython: embedsignature=True

cimport cython
from cython cimport cdivision, boundscheck, wraparound
from libcpp cimport bool as bool_cpp
from libcpp.vector cimport vector
from libcpp.utility cimport pair
from libcpp.algorithm cimport sort
from libc.stdlib cimport malloc, free
from libc.math cimport fabs, INFINITY
import numpy as np
cimport numpy as np
    


@boundscheck(False)
@wraparound(False) 
cpdef bool_cpp is_sorted(double[:] arr):
    cdef long i
    cdef size_t size = arr.shape[0]
    for i in range(size - 1):
        if arr[i] > arr[i + 1]:
            return False
    return True


@boundscheck(False)
@wraparound(False) 
@cdivision(True)  
def compute_voronoi1d_cell_extents(double[:] sites,
                                   double lb=0, 
                                   double ub=INFINITY,
                                   double fill_value=0):
    cdef size_t size = sites.shape[0]
    cdef int i, istart
    cdef double d1, d2
    cdef double[:] extent = np.zeros(size, dtype=np.double)
    size = sites.shape[0]
    extent = np.zeros(size, dtype=np.double)
    if not lb == -INFINITY:
        istart = 0
        d1 = lb
    else:
        istart = 1
        extent[0] = fill_value
        d1 = (sites[0] + sites[1]) / 2.
    for i in range(istart, size - 1):
        d2 = (sites[i] + sites[i+1]) / 2.
        extent[i] = d2 - d1
        d1 = d2
    extent[i+1] = ub - d2 if not ub==INFINITY else fill_value
    return np.asarray(extent)


@boundscheck(False)
@wraparound(False) 
cdef int floor_index(double xp, double[:] x, ssize_t xlen):
    cdef int i
    for i in range(xlen):
        if x[i] <= xp <= x[i+1]:
            return i


@boundscheck(False)
@wraparound(False) 
@cdivision(True)  
cdef double _interpolate_linear_1d(double xp, double[:] x, double[:] y):
    cdef size_t xlen = x.shape[0]
    cdef int i
    cdef double x0, x1, y0, y1
    
    if xp <= x[0]:
        return y[0]
    elif xp >= x[xlen-1]:
        return y[xlen-1]
    i = floor_index(xp, x, xlen)
    x0 = x[i]
    x1 = x[i+1]
    y0 = y[i]
    y1 = y[i+1]
    return y0 + (xp - x0) * (y1 - y0) / (x1 - x0)


@boundscheck(False)
@wraparound(False) 
@cdivision(True)  
cdef double _interpolate_nearest_1d(double xp, double[:] x, double[:] y):
    cdef ssize_t xlen = x.shape[0]
    cdef int i
    cdef double x0, x1, y0, y1
    
    if xp <= x[0]:
        return y[0]
    elif xp >= x[xlen-1]:
        return y[xlen-1]
    i = nearest_neighbour_1d(xp, x, xlen)
    return y[i]


@boundscheck(False)
@wraparound(False) 
cpdef int nearest_neighbour_1d(double xp, double[:] x, ssize_t xlen):
    cdef int i
    cdef double x0, x1
    if xlen==1 or xp<x[0]:
        return 0
    for i in range(xlen - 1):
        x0 = x[i]
        x1 = x[i + 1]
        if x0 <= xp <= x1:
            if fabs(x0-xp) < fabs(x1-xp):
                return i
            return i + 1
    return i + 1


@boundscheck(False)
@wraparound(False) 
cpdef interpolate_depth_profile(double[:] x, double[:] y, double[:] x0):
    cdef size_t ilayer = 0
    cdef size_t i
    cdef size_t x0_len = x0.shape[0]
    cdef size_t x_len = x.shape[0]
    cdef double interface_lower = 0
    cdef double interface_upper = x[0]
    cdef double x_i
    cdef double[:] ynew = np.zeros(x0_len, dtype=np.double)
    
    for i in range(x0_len):
        x_i = x0[i]
        while True:
            if interface_lower <= x_i < interface_upper or \
                ilayer + 1 >= x_len:
                ynew[i] = y[ilayer]
                break
            else:
                ilayer += 1
                interface_lower = interface_upper
                interface_upper += x[ilayer]
    return np.asarray(ynew)


@boundscheck(False)
@wraparound(False) 
cpdef interpolate_linear_1d(xp, double[:] x, double[:] y):
    """ Linear interpolation in 1-D
    Parameters
    ----------
    xp : float, ndarray of floats
        Interpolation point(s)
        
    x, y : ndarray of floats
        Discrete set of known data points
    
    Returns
    -------
    yp : float
        Interpolated value(s)
    """
    if np.isscalar(xp):
        return _interpolate_linear_1d(xp, x, y)
    cdef int i, size = xp.shape[0]
    cdef double[:] yp = np.zeros(size, dtype=np.double)
    for i in range(size):
        yp[i] = _interpolate_linear_1d(xp[i], x, y)
    return np.asarray(yp)


@boundscheck(False)
@wraparound(False) 
cpdef interpolate_nearest_1d(xp, double[:] x, double[:] y):
    """ Nearest-neighbour interpolation in 1-D
    Parameters
    ----------
    xp : float, ndarray of floats
        Interpolation point(s)
        
    x, y : ndarray of floats
        Discrete set of known data points
    
    Returns
    -------
    yp : float
        Interpolated value(s)
    """
    if np.isscalar(xp):
        return _interpolate_nearest_1d(xp, x, y)
    cdef int i, size = xp.shape[0]
    cdef double[:] yp = np.zeros(size, dtype=np.double)
    for i in range(size):
        yp[i] = _interpolate_nearest_1d(xp[i], x, y)
    return np.asarray(yp)


@boundscheck(False)
@wraparound(False) 
@cdivision(True)
cpdef inverse_covariance(double sigma, double r, size_t n):
    cdef size_t i
    cdef double factor = 1 / (sigma**2 * (1 - r**2))
    cdef double[:, ::1] matrix = np.zeros((n, n), dtype=np.double)
    
    for i in range(1, n-1):
        matrix[i, i] = (1 + r**2) * factor
        matrix[i, i-1] = -r * factor
        matrix[i-1, i] = -r * factor
    matrix[0, 0] = factor
    matrix[n-1, n-1] = factor
    matrix[i+1, i] = -r * factor
    matrix[i, i+1] = -r * factor
    return np.asarray(matrix)
    

@boundscheck(False)
@wraparound(False) 
cpdef insert_1d(double[:] values, long index, double value):
    cdef size_t new_size = values.shape[0] + 1
    cdef double[:] new_values = np.zeros(new_size, dtype=np.double)
    cdef size_t i = 0
    cdef size_t j = 0
    while i < new_size:
        if i == index:
            new_values[i] = value
        else:
            new_values[i] = values[j]
            j += 1
        i += 1
    return np.asarray(new_values)


@boundscheck(False)
@wraparound(False) 
cpdef delete_1d(double[:] values, long index):
    cdef size_t size = values.shape[0]
    cdef double[:] new_values = np.zeros(size - 1, dtype=np.double)
    cdef size_t i, j = 0
    for i in range(size):
        if i != index:
            new_values[j] = values[i]
            j += 1
    return np.asarray(new_values)

